
/****************************************************************

  File name   :  math_script_parser.cpp
  Author      :  Ҷ
  Version     :  1.0a
  Create Date :  2014/06/18 
  Description :  ѧű 
  Others      :  

*****************************************************************/

// --------------------------------------------------------------------------------------

#include "misc.h"
#include "operator.h"
#include "common.h"
#include "const_value.h"
#include "value_node_operate.h"
#include "math_expression_parser.h"
#include "math_script_parser.h"
#include <assert.h>

// --------------------------------------------------------------------------------------

#pragma warning(disable: 4996)
namespace Why
{

// --------------------------------------------------------------------------------------

bool MathScriptParser::LoadFile(const char* szFileName)
{
    if (!szFileName || szFileName[0] == 0)
    {
        return false;
    }

    FILE * fp = ::fopen(szFileName, "rb");

    if (NULL == fp)
    {
        return false;
    }

    ::fseek(fp, 0, SEEK_END);

    unsigned int bufSize = ::ftell(fp);
    if (bufSize == 0)
    {
        ::fclose(fp);
        return false;
    }

    ::fseek(fp, 0, SEEK_SET);

    char* bufPtr = (char* )malloc(bufSize + 1);
    assert(bufPtr != NULL);

    if (::fread(bufPtr, sizeof(char), bufSize, fp) != bufSize)
    {
        ::fclose(fp);
        free(bufPtr);
        return false;
    }

    ::fclose(fp);

    Clear();

    m_scriptBufPtr = bufPtr;
    m_scriptBufPtr[bufSize] = 0;
    m_scriptBufSize = bufSize;

    return Parse();
}

bool MathScriptParser::ProcessMemory(const char* bufPtr, unsigned int bufSize)
{
    if (!bufPtr || bufSize == 0)
    {
        return false;
    }

    Clear();

    m_scriptBufPtr = (char* )malloc(bufSize + 1);
    memcpy(m_scriptBufPtr, bufPtr, bufSize);
    m_scriptBufPtr[bufSize] = 0;
    m_scriptBufSize = bufSize;

    return Parse();
}

bool MathScriptParser::Parse()
{
    m_vecScriptLines.reserve(32);

    char* buffer = (char*)m_scriptBufPtr;
    unsigned int count = 0;
    unsigned int i;
    for (i = 0; i < m_scriptBufSize; i++)
    {
        if ( (buffer[i] == '\r') || (buffer[i] == '\n') || (buffer[i] == 0) || (buffer[i] == ';'))
        {
            buffer[i] = 0;
            count = 0;
        }
        else
        {
            if (count == 0)
            {
                m_vecScriptLines.push_back(&buffer[i]);
            }
            count++;
        }
    }

    size_t numLines = m_vecScriptLines.size();
    for (size_t i = 0; i < numLines; i++)
    {
        if (!ParseLine(i))
        {
            return false;
        }
    }

    return true;
}

bool MathScriptParser::ParseLine(unsigned int line)
{
    m_currentLine = line;

    char szInfo[256];
    char szWord[64];

    const char* szExpression = m_vecScriptLines[m_currentLine];
    unsigned int pos = 0;

    // ҵһ
    if (!GetWord(szWord, szExpression, pos))
    {
        sprintf(szInfo, "[Warning] Parse Failed: Line:%d, Expression:%s\r\n", m_currentLine, szExpression);
        strcat(m_szParseInfo, szInfo);
        return true;
    }

    // ҵȺ
    if (!GetEqualOperator(szExpression, pos))
    {
        sprintf(szInfo, "[Warning] Parse Failed: Line:%d, Expression:%s\r\n", m_currentLine, szExpression);
        strcat(m_szParseInfo, szInfo);
        return false;
    }

    size_t lenWord = strlen(szWord);
    if (lenWord != 1)
    {
        if (stricmp(szWord, "vertices") == 0)
        {
            ParseVerticesCount(szExpression + pos);
        }
        else
        {
            sprintf(szInfo, "[Warning] Parse Failed: Line:%d, Word:%s, Expression:%s\r\n", m_currentLine, szWord, szExpression);
            strcat(m_szParseInfo, szInfo);
        }

        return true;
    }
    
    // ñ
    if (szWord[0] < 'a' || szWord[0] > 'z')
    {
        // ûбֵı
        sprintf(szInfo, "[Warning] Parse Failed: Line:%d, Word:%s, Expression:%s\r\n", m_currentLine, szWord, szExpression);
        strcat(m_szParseInfo, szInfo);
        return false;
    }

    if (m_nVerticesCount == 0)
    {
        // ûöĿ
        sprintf(szInfo, "[Warning] Vertices count is not setted: Line:%d, Word:%s, Expression:%s\r\n", m_currentLine, szWord, szExpression);
        strcat(m_szParseInfo, szInfo);
        return false;
    }

    ValueNode& rst = m_listValueNode[szWord[0] - 'a'];

    if (ParseFromToSetting(szExpression + pos, rst))
    {
        return true;
    }

    if (!ParseExpression(szExpression + pos, rst))
    {
        sprintf(szInfo, "[Warning] ParseExpression Failed. Line:%d, Expression:%s\r\n", m_currentLine, szExpression);
        strcat(m_szParseInfo, szInfo);
    }

    return true;
}

bool MathScriptParser::ParseVerticesCount(const char* szExpression)
{
    char szInfo[256];
    char szWord[MAX_WORD_LENGTH];
    unsigned int pos = 0;

    // Ŀ
    if (m_nVerticesCount != 0)
    {
        // Ŀظ
        sprintf(szInfo, "[Warning] Vertices count should be set once. Line:%d, Expression:%s\r\n", m_currentLine, szExpression);
        strcat(m_szParseInfo, szInfo);
    }

    if (!GetWord(szWord, szExpression, pos))
    {
        sprintf(szInfo, "[Warning] Failed to set vertices count. Line:%d, Expression:%s\r\n", m_currentLine, szExpression);
        strcat(m_szParseInfo, szInfo);
    }

    float v;
    if (GetIntegerValue(szWord, v))
    {
        m_nVerticesCount = (unsigned int)(v + 0.1f);
        m_nVerticesCountInDimension1 = 0;
        m_nVerticesCountInDimension2 = 0;
        return true;
    }

    bool sus = false;

    // vertices = dimension1:100 dimension2:100
    // vertices = D1:100 D2:100

    int n;
    if (stricmp(szWord, "dimension1") == 0 || stricmp(szWord, "d1") == 0)
    {
        if (GetColonOperator(szExpression, pos))
        {
            if (GetIntegerValue(n, szExpression, pos))
            {
                m_nVerticesCountInDimension1 = (unsigned int)n;

                if (GetWord(szWord, szExpression, pos))
                {
                    if (stricmp(szWord, "dimension2") == 0 || stricmp(szWord, "d2") == 0)
                    {
                        if (GetColonOperator(szExpression, pos))
                        {
                            if (GetIntegerValue(n, szExpression, pos))
                            {
                                m_nVerticesCountInDimension2 = (unsigned int)n;

                                if (m_nVerticesCountInDimension1 > 0 &&
                                    m_nVerticesCountInDimension2 > 0)
                                {
                                    m_nVerticesCount = m_nVerticesCountInDimension1 * m_nVerticesCountInDimension2;
                                    sus = true;
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    if (!sus)
    {
        m_nVerticesCount = 0;
        m_nVerticesCountInDimension1 = 0;
        m_nVerticesCountInDimension2 = 0;
        sprintf(szInfo, "[Warning] Failed to set vertices count. Line:%d, Expression:%s\r\n", m_currentLine, szExpression);
        strcat(m_szParseInfo, szInfo);
    }
    return sus;
}

bool MathScriptParser::ParseFromToSetting(const char* szExpression, ValueNode& rst)
{
    char szInfo[256];
    char szWord[MAX_WORD_LENGTH];
    unsigned int pos = 0;

    if (!GetWord(szWord, szExpression, pos))
    {
        return false;
    }

    if (stricmp(szWord, "from") != 0)
    {
        return false;
    }

    float vs, ve;
    bool ableValue = false;

    unsigned int bracketIndex;
    if (GetLeftBracketOperator(szExpression, pos, bracketIndex))
    {
        ableValue = Why::ParseBracketExpression(szExpression, pos, vs, bracketIndex, false);
    }
    else
    {
        ableValue = GetScientificFloat(vs, szExpression, pos);
        if (!ableValue)
        {
            bool nv = false;
            if (szExpression[pos] == '-')
            {
                nv = true;
                pos++;
            }
            else if (szExpression[pos] == '+')
            {
                pos++;
            }

            if (GetWord(szWord, szExpression, pos))
            {
                ableValue = GetConstValue(szWord, vs);
                if (!ableValue)
                {
                    ableValue = GetIntegerValue(szWord, vs);
                }
                if (!ableValue)
                {
                    ableValue = GetFloatValue(szWord, vs);
                }
            }

            if (nv)
            {
                vs = -vs;
            }
        }
    }

    if (!ableValue)
    {
        sprintf(szInfo, "[Warning] Line:%d ParseFromToSetting Fail to get the \"from\" value:%s\r\n", m_currentLine, szExpression);
        strcat(m_szParseInfo, szInfo);
        return false;
    }

    if (!GetWord(szWord, szExpression, pos))
    {
        sprintf(szInfo, "[Warning] Line:%d ParseFromToSetting Fail to get the \"to\" word:%s\r\n", m_currentLine, szExpression);
        strcat(m_szParseInfo, szInfo);
        return false;
    }

    if (stricmp(szWord, "to") != 0)
    {
        sprintf(szInfo, "[Warning] Line:%d ParseFromToSetting Fail to get the \"to\" word:%s\r\n", m_currentLine, szExpression);
        strcat(m_szParseInfo, szInfo);
        return false;
    }

    ableValue = false;
    if (GetLeftBracketOperator(szExpression, pos, bracketIndex))
    {
        ableValue = Why::ParseBracketExpression(szExpression, pos, ve, bracketIndex, false);
    }
    else
    {
        ableValue = GetScientificFloat(ve, szExpression, pos);
        if (!ableValue)
        {
            bool nv = false;
            if (szExpression[pos] == '-')
            {
                nv = true;
                pos++;
            }
            else if (szExpression[pos] == '+')
            {
                pos++;
            }

            if (GetWord(szWord, szExpression, pos))
            {
                ableValue = GetConstValue(szWord, ve);
                if (!ableValue)
                {
                    ableValue = GetIntegerValue(szWord, ve);
                }
                if (!ableValue)
                {
                    ableValue = GetFloatValue(szWord, ve);
                }
            }

            if (nv)
            {
                ve = -ve;
            }
        }
    }

    if (!ableValue)
    {
        sprintf(szInfo, "[Warning] Line:%d ParseFromToSetting Fail to get the \"to\" value:%s\r\n", m_currentLine, szExpression);
        strcat(m_szParseInfo, szInfo);
        return false;
    }

    if (!IsValue2D())
    {
        rst.Resize(m_nVerticesCount);
        rst.LinerInterpolate(vs, ve);
    }
    else
    {
        if (!GetWord(szWord, szExpression, pos))
        {
            sprintf(szInfo, "[Warning] Line:%d ParseFromToSetting not set dimension:%s\r\n", m_currentLine, szExpression);
            strcat(m_szParseInfo, szInfo);
            return false;
        }

        if (stricmp(szWord, "dimension1") == 0 || stricmp(szWord, "d1") == 0)
        {
            rst.Resize(m_nVerticesCountInDimension1);
            rst.LinerInterpolate(vs, ve);
            rst.WholeDuplicate(m_nVerticesCountInDimension2);
        }
        else if (stricmp(szWord, "dimension2") == 0 || stricmp(szWord, "d2") == 0)
        {
            rst.Resize(m_nVerticesCountInDimension2);
            rst.LinerInterpolate(vs, ve);
            rst.SingleDuplicate(m_nVerticesCountInDimension1);
        }
        else
        {
            sprintf(szInfo, "[Warning] Line:%d ParseFromToSetting not set dimension:%s\r\n", m_currentLine, szExpression);
            strcat(m_szParseInfo, szInfo);
            return false;
        }
    }

    return true;
}

bool MathScriptParser::ParseExpression(const char* szExpression, ValueNode& rst)
{
    if (!szExpression || !szExpression[0])
    {
        return false;
    }

    unsigned int lenExp = strlen(szExpression);
    unsigned int pos = 0;

    // ֵ
    ValueNode listValues[3];
    unsigned int valuesCount = 0;

    // ׼Ŀ
    const Operator* standardSingleOperator = NULL;
    standardSingleOperator = GetStandardSingleOperator(szExpression, pos);

    // ׼˫Ŀ
    const Operator* standardTwinOperators[2];
    memset(standardTwinOperators, 0, sizeof(standardTwinOperators));
    unsigned int operatorsCount = 0;

    char szWord[MAX_WORD_LENGTH];

    while (pos < lenExp)
    {
        if (valuesCount > operatorsCount)
        {
            // ֵĿĿʱ,ñ׼˫Ŀ

            const Operator* operatorPtr = GetStandardTwinOperator(szExpression, pos);
            if (!operatorPtr)
            {
                if (pos >= lenExp)
                {
                    // ʽĸǿո,Ҫ
                    break;
                }
                else
                {
                    return false;
                }
            }

            if (operatorsCount == 1)
            {
                if (operatorPtr->szName[0] == '+' || operatorPtr->szName[0] == '-')
                {
                    // ڶΪ+-ʱ,һԷ
                    ValueNodeTwinOperate(
                        standardTwinOperators[0]->twin_func, 
                        &listValues[0], 
                        &listValues[0], 
                        &listValues[1]);
                    standardTwinOperators[0] = operatorPtr;
                    valuesCount--;
                }
                else
                {
                    standardTwinOperators[operatorsCount] = operatorPtr;
                    operatorsCount++;
                }
            }
            else
            {
                standardTwinOperators[operatorsCount] = operatorPtr;
                operatorsCount++;
            }
        }
        else
        {
            // ֵ

            bool ableValue = false;
            ValueNode valueNode;

            // 
            {
                unsigned int bracketIndex;
                const Operator* operatorPtr = GetLeftBracketOperator(szExpression, pos, bracketIndex);
                if (operatorPtr)
                {
                    if (!ParseBracketExpression(szExpression, pos, valueNode, bracketIndex, false))
                    {
                        return false;
                    }
                    ableValue = true;
                }
            }

            // ѧʽĸ
            if (!ableValue)
            {
                ableValue = GetScientificFloat(valueNode._value, szExpression, pos);
            }

            // ȡǰ
            if (!ableValue && !GetWord(szWord, szExpression, pos))
            {
                return false;
            }

            // 
            if (!ableValue)
            {
                ableValue = GetConstValue(szWord, valueNode._value);
            }

            // ֵ
            if (!ableValue)
            {
                ableValue = GetIntegerValue(szWord, valueNode._value);
            }

            // ֵ
            if (!ableValue)
            {
                ableValue = GetFloatValue(szWord, valueNode._value);
            }

            // ֵ
            if (!ableValue)
            {
                char _param = szWord[1] == 0 ? szWord[0] : 0;
                if (_param)
                {
                    valueNode.Copy(m_listValueNode[_param - 'a']);
                    ableValue = true;
                }
            }

            // 
            if (!ableValue)
            {
                const Operator* operatorPtr = GetFunctionOperator(szWord);
                if (operatorPtr)
                {
                    if (!ParseOperatorExpression(szExpression, pos, operatorPtr, valueNode))
                    {
                        return false;
                    }
                    ableValue = true;
                }
            }

            if (!ableValue)
            {
                return false;
            }

            if (operatorsCount == 2)
            {
                assert(valuesCount == 2);

                listValues[valuesCount].Copy(valueNode);

                const Operator* operatorPtr = standardTwinOperators[1];
                ValueNodeTwinOperate(
                    operatorPtr->twin_func, 
                    &listValues[1], 
                    &listValues[1], 
                    &listValues[2]);
                operatorsCount--;
            }
            else if (operatorsCount == 1)
            {
                assert(valuesCount == 1);

                listValues[valuesCount].Copy(valueNode);

                const Operator* operatorPtr = standardTwinOperators[0];
                if (operatorPtr->szName[0] != '+' && operatorPtr->szName[0] != '-')
                {
                    ValueNodeTwinOperate(
                        operatorPtr->twin_func, 
                        &listValues[0], 
                        &listValues[0], 
                        &listValues[1]);
                    operatorsCount--;
                }
                else
                {
                    valuesCount++;
                }
            }
            else if (operatorsCount == 0)
            {
                assert(valuesCount == 0);

                listValues[valuesCount].Copy(valueNode);

                if (standardSingleOperator)
                {
                    ValueNodeSingleOperate(
                        standardSingleOperator->single_func, 
                        &listValues[0], 
                        &listValues[0]);
                    standardSingleOperator = NULL;
                }

                valuesCount++;
            }
            else
            {
                assert(0);
            }
        }
    }

    if (valuesCount != operatorsCount + 1)
    {
        return false;
    }

    if (operatorsCount == 1)
    {
        const Operator* operatorPtr = standardTwinOperators[0];
        ValueNodeTwinOperate(
            operatorPtr->twin_func, 
            &listValues[0], 
            &listValues[0], 
            &listValues[1]);
        operatorsCount--;
        valuesCount--;
    }

    rst.Copy(listValues[0]);

    return true;
}

// ڵӱʽ
bool MathScriptParser::ParseBracketExpression(
    const char* szExpression, unsigned int& pos, ValueNode& rst, 
    unsigned int bracketIndex, bool endComma)
{
    rst.Clear();

    int nLeftBracket = 1;
    unsigned int bracketIndices[MAX_BRACKET_SIZE];
    bracketIndices[0] = bracketIndex;

    char szSubExp[SUB_EXPRESSION_MAX_LENGTH];
    memset(szSubExp, 0, SUB_EXPRESSION_MAX_LENGTH);

    unsigned int leftBI;
    unsigned int rightBI;

    unsigned int subExpLen = 0;
    while (szExpression[pos + subExpLen])
    {
        leftBI = IsLeftBracketLetter(szExpression[pos + subExpLen]);
        rightBI = IsRightBracketLetter(szExpression[pos + subExpLen]);

        if (leftBI != -1)
        {
            if (nLeftBracket > MAX_BRACKET_SIZE - 1)
            {
                return false;
            }
            bracketIndices[nLeftBracket] = leftBI;
            nLeftBracket++;
        }
        else if (rightBI != -1)
        {
            nLeftBracket--;
            if (bracketIndices[nLeftBracket] != rightBI)
            {
                return false;
            }
        }
        else if (endComma && szExpression[pos + subExpLen] == ',')
        {
            // Զβ
            if (nLeftBracket == 1)
            {
                nLeftBracket--;
            }
            else
            {
                return false;
            }
        }

        if (nLeftBracket == 0)
        {
            break;
        }
        szSubExp[subExpLen] = szExpression[pos + subExpLen];
        subExpLen++;
    }

    if (nLeftBracket != 0)
    {
        return false;
    }

    pos += subExpLen + 1;

    return ParseExpression(szSubExp, rst);
}

// ӱʽ
bool MathScriptParser::ParseOperatorExpression(const char* szExpression, unsigned int& pos, 
    const Operator* operatorPtr, ValueNode& rst)
{
    switch (operatorPtr->type)
    {
    case OT_STANDARD_SINGLE:
    case OT_SINGLE:
        return ParseSingleOperatorExpression(szExpression, pos, operatorPtr->single_func, rst);
        break;
    case OT_STANDARD_TWIN:
    case OT_TWIN:
        return ParseTwinOperatorExpression(szExpression, pos, operatorPtr->twin_func, rst);
        break;
    case OT_THREE:
        return ParseThreeOperatorExpression(szExpression, pos, operatorPtr->three_func, rst);
        break;
    case OT_FOUR:
        return ParseFourOperatorExpression(szExpression, pos, operatorPtr->four_func, rst);
        break;
    case OT_VALUENODE_TO_FLOAT:
        return ParseValueNodeToFloatOperatorExpression(szExpression, pos, operatorPtr->value_node_to_float_func, rst);
        break;
    case OT_VALUENODE_CONVERT:
        return ParseValueNodeConvertOperatorExpression(szExpression, pos, operatorPtr->value_node_convert_func, rst);
        break;
    default:
        return false;
    }
}

// ڵӱʽ
bool MathScriptParser::ParseValueNodeToFloatOperatorExpression(const char* szExpression, unsigned int& pos, 
    VALUENODE_TO_FLOAT_OPERATOR_FUNC func, ValueNode& rst)
{
    if (!func)
    {
        assert(0);
        return false;
    }

    unsigned int bracketIndex;
    if (!GetLeftBracketOperator(szExpression, pos, bracketIndex))
    {
        return false;
    }

    ValueNode value0;
    if (!ParseBracketExpression(szExpression, pos, value0, bracketIndex, false))
    {
        return false;
    }

    float v = func(&value0);
    rst.SetValue(v);

    return true;
}

// ڵӱʽ
bool MathScriptParser::ParseValueNodeConvertOperatorExpression(const char* szExpression, unsigned int& pos, 
    VALUENODE_CONVERT_OPERATOR_FUNC func, ValueNode& rst)
{
    if (!func)
    {
        assert(0);
        return false;
    }

    unsigned int bracketIndex;
    if (!GetLeftBracketOperator(szExpression, pos, bracketIndex))
    {
        return false;
    }

    ValueNode value0;
    if (!ParseBracketExpression(szExpression, pos, value0, bracketIndex, false))
    {
        return false;
    }

    func(&value0, &rst);

    return true;
}

// Ŀӱʽ
bool MathScriptParser::ParseSingleOperatorExpression(const char* szExpression, unsigned int& pos, 
    SINGLE_OPERATOR_FUNC func, ValueNode& rst)
{
    if (!func)
    {
        assert(0);
        return false;
    }

    unsigned int bracketIndex;
    if (!GetLeftBracketOperator(szExpression, pos, bracketIndex))
    {
        return false;
    }

    ValueNode value0;
    if (!ParseBracketExpression(szExpression, pos, value0, bracketIndex, false))
    {
        return false;
    }

    ValueNodeSingleOperate(func, &rst, &value0);

    return true;
}

// ˫Ŀӱʽ
bool MathScriptParser::ParseTwinOperatorExpression(const char* szExpression, unsigned int& pos, 
    TWIN_OPERATOR_FUNC func, ValueNode& rst)
{
    if (!func)
    {
        assert(0);
        return false;
    }

    unsigned int bracketIndex;
    if (!GetLeftBracketOperator(szExpression, pos, bracketIndex))
    {
        return false;
    }

    ValueNode value0;
    if (!ParseBracketExpression(szExpression, pos, value0, bracketIndex, true))
    {
        return false;
    }

    ValueNode value1;
    if (!ParseBracketExpression(szExpression, pos, value1, bracketIndex, false))
    {
        return false;
    }

    ValueNodeTwinOperate(func, &rst, &value0, &value1);

    return true;
}

// Ŀӱʽ
bool MathScriptParser::ParseThreeOperatorExpression(const char* szExpression, unsigned int& pos, 
    THREE_OPERATOR_FUNC func, ValueNode& rst)
{
    if (!func)
    {
        assert(0);
        return false;
    }

    unsigned int bracketIndex;
    if (!GetLeftBracketOperator(szExpression, pos, bracketIndex))
    {
        return false;
    }

    ValueNode value0;
    if (!ParseBracketExpression(szExpression, pos, value0, bracketIndex, true))
    {
        return false;
    }

    ValueNode value1;
    if (!ParseBracketExpression(szExpression, pos, value1, bracketIndex, true))
    {
        return false;
    }

    ValueNode value2;
    if (!ParseBracketExpression(szExpression, pos, value2, bracketIndex, false))
    {
        return false;
    }

    ValueNodeThreeOperate(func, &rst, &value0, &value1, &value2);

    return true;
}

// Ŀӱʽ
bool MathScriptParser::ParseFourOperatorExpression(const char* szExpression, unsigned int& pos, 
    FOUR_OPERATOR_FUNC func, ValueNode& rst)
{
    if (!func)
    {
        assert(0);
        return false;
    }

    unsigned int bracketIndex;
    if (!GetLeftBracketOperator(szExpression, pos, bracketIndex))
    {
        return false;
    }

    ValueNode value0;
    if (!ParseBracketExpression(szExpression, pos, value0, bracketIndex, true))
    {
        return false;
    }

    ValueNode value1;
    if (!ParseBracketExpression(szExpression, pos, value1, bracketIndex, true))
    {
        return false;
    }

    ValueNode value2;
    if (!ParseBracketExpression(szExpression, pos, value2, bracketIndex, true))
    {
        return false;
    }

    ValueNode value3;
    if (!ParseBracketExpression(szExpression, pos, value3, bracketIndex, false))
    {
        return false;
    }

    ValueNodeFourOperate(func, &rst, &value0, &value1, &value2, &value3);

    return true;
}

void MathScriptParser::Clear()
{
    for (unsigned int i = 0; i < 26; i++)
    {
        m_listValueNode[i].Clear();
    }

    m_nVerticesCount = 0;
    m_nVerticesCountInDimension1 = 0;
    m_nVerticesCountInDimension2 = 0;
    memset(m_szParseInfo, 0, sizeof(m_szParseInfo));

    if (m_scriptBufPtr)
    {
        free(m_scriptBufPtr);
        m_scriptBufPtr = NULL;
    }
    m_scriptBufSize = 0;

    m_vecScriptLines.clear();
}

bool MathScriptParser::Reparse()
{
    for (unsigned int i = 0; i < 26; i++)
    {
        m_listValueNode[i].Clear();
    }

    m_nVerticesCount = 0;
    m_nVerticesCountInDimension1 = 0;
    m_nVerticesCountInDimension2 = 0;
    memset(m_szParseInfo, 0, sizeof(m_szParseInfo));

    m_vecScriptLines.clear();

    return Parse();
}

unsigned int MathScriptParser::GetVerticesCount() const
{
    return m_nVerticesCount;
}

// ǷΪ2D
bool MathScriptParser::IsValue2D() const
{
    return (m_nVerticesCountInDimension1 > 0 && m_nVerticesCountInDimension2 > 0);
}

unsigned int MathScriptParser::GetVerticesCountInDimension1() const
{
    return m_nVerticesCountInDimension1;
}

unsigned int MathScriptParser::GetVerticesCountInDimension2() const
{
    return m_nVerticesCountInDimension2;
}

const ValueNode* MathScriptParser::GetValueNode(char c) const
{
    if (c >= 'a' && c <= 'z')
    {
        int index = c - 'a';
        return m_listValueNode + index;
    }
    else
    {
        return NULL;
    }
}

unsigned int MathScriptParser::GetScriptLinesCount() const
{
    return m_vecScriptLines.size();
}

const char* MathScriptParser::GetScriptLine(unsigned int index) const
{
    if (index >= m_vecScriptLines.size())
    {
        return NULL;
    }
    else
    {
        return m_vecScriptLines[index];
    }
}

const char* MathScriptParser::GetParseInfo() const
{
    return m_szParseInfo;
}

unsigned int MathScriptParser::GetScriptBufSize() const
{
    return m_scriptBufSize;
}

void MathScriptParser::GetScriptBuffer(char* szBuffer, unsigned int bufSize) const
{
    szBuffer[0] = 0;
    unsigned int usedSize = 1;

    size_t numLines = m_vecScriptLines.size();
    for (size_t i = 0; i < numLines; i++)
    {
        unsigned int s = strlen(m_vecScriptLines[i]);
        if (usedSize + s + 2 > bufSize)
        {
            break;
        }

        strcat(szBuffer, m_vecScriptLines[i]);
        strcat(szBuffer, "\r\n");

        usedSize += s;
        usedSize += 2;
    }
}

bool MathScriptParser::IsReadyXYZ() const
{
    if (m_nVerticesCount == 0)
    {
        return false;
    }

    return (m_listValueNode['x' - 'a']._arrayValues ||
            m_listValueNode['y' - 'a']._arrayValues ||
            m_listValueNode['z' - 'a']._arrayValues);
}

bool MathScriptParser::IsValidRGB() const
{
    if (m_nVerticesCount == 0)
    {
        return false;
    }

    return (m_listValueNode['r' - 'a']._arrayValues &&
            m_listValueNode['g' - 'a']._arrayValues &&
            m_listValueNode['b' - 'a']._arrayValues);
}

//ڵuvǷЧ(Чʱŷtrue)
bool MathScriptParser::IsValidUV() const
{
    if (m_nVerticesCount == 0)
    {
        return false;
    }

    return (m_listValueNode['u' - 'a']._arrayValues ||
            m_listValueNode['v' - 'a']._arrayValues);
}

};