
/****************************************************************

  File name   :  Operator.cpp
  Author      :  Ҷ
  Version     :  1.0a
  Create Date :  2014/06/18 
  Description :   
  Others      :  

*****************************************************************/

// --------------------------------------------------------------------------------------

#include <cstdlib>
#include <cmath>
#include <cfloat>
#include <cstdio>
#include <assert.h>

#include "misc.h"
#include "value_node.h"
#include "operator.h"

// --------------------------------------------------------------------------------------

#pragma warning(disable: 4996)
namespace Why
{

// --------------------------------------------------------------------------------------

// 

static float     yf_positive(float a)
{
    return a;
}

static float     yf_negative(float a)
{
    return -a;
}

static float     yf_add(float a, float b)
{
    return a + b;
}

static float     yf_sub(float a, float b)
{
    return a - b;
}

static float     yf_multiply(float a, float b)
{
    return a * b;
}

static float     yf_divide(float a, float b)
{
    if (b > -FLT_EPSILON && b < FLT_EPSILON)
    {
        return a > 0 ? FLT_MAX : -FLT_MAX;
    }
    else
    {
        return a / b;
    }
}

static float     yf_abs(float a)
{
    return ::fabsf(a);
}

static float     yf_floor(float a)
{
    return ::floorf(a);
}

static float     yf_ceil(float a)
{
    return ::ceilf(a);
}

static float     yf_sign(float a)
{
    return (a >= 0.0f) ? 1.0f : -1.0f;
}

static float     yf_sqrt(float a)
{
    if (a < 0.0f)
    {
        return 0.0f;
    }
    else
    {
        return ::sqrtf(a);
    }
}

static float     yf_exp(float a)
{
    /*
    if (a < 0.0f)
    {
        return 1.0f;
    }
    else
    {
        return ::expf(a);
    }
    */
    return ::expf(a);
}

static float     yf_log(float a)
{
    if (a < 0.0f)
    {
        return 0.0f;
    }
    else
    {
        return ::logf(a);
    }
}

static float     yf_log2(float a)
{
    if (a < 0.0f)
    {
        return 0.0f;
    }
    else
    {
        const float ln2 = (float)0.693147180559945309417;
        return ::logf(a) / ln2;
    }
}

static float     yf_log10(float a)
{
    if (a < 0.0f)
    {
        return 0.0f;
    }
    else
    {
        return ::log10f(a);
    }
}

static float     yf_sin(float a)
{
    return ::sinf(a);
}

static float     yf_cos(float a)
{
    return ::cosf(a);
}

static float     yf_tan(float a)
{
    return ::tanf(a);
}

static float     yf_asin(float f)
{
    if(f >= 1.0f)   return REAL_PI*0.5f;
    if(f <= -1.0f)  return -REAL_PI*0.5f;
    return ::asinf(f);
}

static float     yf_acos(float f)
{
    if(f >= 1.0f)   return 0.0f;
    if(f <= -1.0f)  return REAL_PI;
    return ::acosf(f);
}

static float     yf_atan(float a)
{
    return ::atanf(a);
}

static float     yf_rand(float a)
{
    return a * (float)(::rand() / ((float)RAND_MAX + 1));
}

// --------------------------------------------------------------------------------------

static float     yf_max(float a, float b)
{
    return (a < b) ? b : a;
}

static float     yf_min(float a, float b)
{
    return (a < b) ? a : b;
}

static float     yf_mod(float a, float b)
{
    if (b < FLT_EPSILON)
    {
        return 0.0f;
    }
    else
    {
        return ::fmodf(a, b);
    }
}

static float     yf_pow(float a, float b)
{
    return ::powf(a, b);
}

static float     yf_atan2(float x, float y)
{
    return ::atan2f(x, y);
}

static float     yf_rand2(float a, float b)
{
    const float r = (float)(::rand() / ((float)RAND_MAX + 1));
    return r*(b-a) + a;
}

// --------------------------------------------------------------------------------------

static float     yf_lerp(float a, float b, float r)
{
    return a + (b - a) * r;
}

static float     yf_clamp(float v, float hi, float low)
{
    if (v > hi) 
        return hi;
    else if (v < low) 
        return low;
    else
        return v;
}

static float     yf_limit(float v, float _min, float _max)
{
    if (v < _min)
    {
        v = _min;
    }
    else if (v > _max)
    {
        v = _max;
    }
    return v;
}

static float     yf_gray(float r, float g, float b)
{
    return r*0.299f + g*0.587f + b*0.114f;
}

static float     yf_add3(float a, float b, float c)
{
    return a+b+c;
}

static float     yf_min3(float a, float b, float c)
{
    float r = (a < b) ? a : b;
    r = (c < r) ? c : r;
    return r;
}

static float     yf_max3(float a, float b, float c)
{
    float r = (a > b) ? a : b;
    r = (c > r) ? c : r;
    return r;
}

static float     yf_average3(float a, float b, float c)
{
    return (a+b+c)/3.0f;
}

// --------------------------------------------------------------------------------------

static float     yf_add4(float a, float b, float c, float d)
{
    return a+b+c+d;
}

static float     yf_min4(float a, float b, float c, float d)
{
    float r = (a < b) ? a : b;
    r = (c < r) ? c : r;
    r = (d < r) ? d : r;
    return r;
}

static float     yf_max4(float a, float b, float c, float d)
{
    float r = (a > b) ? a : b;
    r = (c > r) ? c : r;
    r = (d > r) ? d : r;
    return r;
}

static float     yf_average4(float a, float b, float c, float d)
{
    return (a+b+c+d)*0.25f;
}

// --------------------------------------------------------------------------------------

static float     yf_add_array(const ValueNode* p)
{
    if (0 == p->_arrayValues)
    {
        return p->_value;
    }

    float a = 0;
    for (unsigned int i = 0; i < p->_valueCount; i++)
    {
        a += p->_arrayValues[i];
    }
    return a;
}

static float     yf_min_array(const ValueNode* p)
{
    if (0 == p->_arrayValues)
    {
        return p->_value;
    }

    float a = p->_arrayValues[0];
    for (unsigned int i = 1; i < p->_valueCount; i++)
    {
        if (a > p->_arrayValues[i])
        {
            a = p->_arrayValues[i];
        }
    }
    return a;
}

static float     yf_max_array(const ValueNode* p)
{
    if (0 == p->_arrayValues)
    {
        return p->_value;
    }

    float a = p->_arrayValues[0];
    for (unsigned int i = 1; i < p->_valueCount; i++)
    {
        if (a < p->_arrayValues[i])
        {
            a = p->_arrayValues[i];
        }
    }
    return a;
}

static float     yf_average_array(const ValueNode* p)
{
    if (0 == p->_arrayValues)
    {
        return p->_value;
    }

    float a = 0;
    for (unsigned int i = 0; i < p->_valueCount; i++)
    {
        a += p->_arrayValues[i];
    }
    return a / p->_valueCount;
}

// --------------------------------------------------------------------------------------

// 
static void     yf_array_move_right(const ValueNode* p, ValueNode* pOut)
{
    if (0 == p->_arrayValues)
    {
        pOut->SetValue(p->_value);
        return;
    }

    pOut->Resize(p->_valueCount);
    for (unsigned int i = p->_valueCount - 1; i > 0; i--)
    {
        pOut->_arrayValues[i] = p->_arrayValues[i - 1];
    }
}

// 
static void     yf_array_move_left(const ValueNode* p, ValueNode* pOut)
{
    if (0 == p->_arrayValues)
    {
        pOut->SetValue(p->_value);
        return;
    }

    pOut->Resize(p->_valueCount);
    for (unsigned int i = 0; i < p->_valueCount - 1; i++)
    {
        pOut->_arrayValues[i] = p->_arrayValues[i + 1];
    }
}

// ۼ
static void     yf_array_cumulate(const ValueNode* p, ValueNode* pOut)
{
    if (0 == p->_arrayValues)
    {
        pOut->SetValue(p->_value);
        return;
    }

    pOut->Resize(p->_valueCount);
    for (unsigned int i = 1; i < p->_valueCount; i++)
    {
        pOut->_arrayValues[i] += p->_arrayValues[i - 1];
    }
}

// --------------------------------------------------------------------------------------

const Operator g_standardSingleOperators[STANDARD_SINGLE_OPERATORS_COUNT] =
{
    {"+",       1, OT_STANDARD_SINGLE, yf_positive, 0, 0, 0, 0, 0},
    {"-",       1, OT_STANDARD_SINGLE, yf_negative, 0, 0, 0, 0, 0},
};

const Operator g_standardTwinOperators[STANDARD_TWIN_OPERATORS_COUNT] =
{
    {"+",       1, OT_STANDARD_TWIN, 0, yf_add, 0, 0, 0, 0},
    {"-",       1, OT_STANDARD_TWIN, 0, yf_sub, 0, 0, 0, 0},
    {"*",       1, OT_STANDARD_TWIN, 0, yf_multiply, 0, 0, 0, 0},
    {"/",       1, OT_STANDARD_TWIN, 0, yf_divide, 0, 0, 0, 0},
    {"%",       1, OT_STANDARD_TWIN, 0, yf_mod, 0, 0, 0, 0},
    {"^",       1, OT_STANDARD_TWIN, 0, yf_pow, 0, 0, 0, 0},
};

const Operator g_leftBracketOperator[BRACKET_OPERATORS_COUNT] = 
{
    {"(",       1, OT_LEFT_BRACKET, 0, 0, 0, 0, 0, 0},
    {"[",       1, OT_LEFT_BRACKET, 0, 0, 0, 0, 0, 0},
    {"}",       1, OT_LEFT_BRACKET, 0, 0, 0, 0, 0, 0},
};

const Operator g_rightBracketOperator[BRACKET_OPERATORS_COUNT] =  
{
    {")",       1, OT_RIGHT_BRACKET, 0, 0, 0, 0, 0, 0},
    {"]",       1, OT_RIGHT_BRACKET, 0, 0, 0, 0, 0, 0},
    {"}",       1, OT_RIGHT_BRACKET, 0, 0, 0, 0, 0, 0},
};

// --------------------------------------------------------------------------------------

// 㺯
Operator       g_arrayFunctionOperators[MAX_FUNCTIONS_COUNT] = {0};
unsigned int   g_countFunctionOperators = 0;

#define SINGLE_OPERATORS_COUNT              18         
#define TWIN_OPERATORS_COUNT                10      
#define THREE_OPERATORS_COUNT               8         
#define FOUR_OPERATORS_COUNT                4            
#define VALUENODE_TO_FLOAT_OPERATORS_COUNT  4            
#define VALUENODE_CONVERT_OPERATORS_COUNT   3  

// ʼ㺯
void InitFunctionOperators()
{
    g_countFunctionOperators = 0;

    const Operator singleOperators[SINGLE_OPERATORS_COUNT] =
    {
        {"positive",0, OT_SINGLE, yf_positive,  0, 0, 0, 0, 0},
        {"negative",0, OT_SINGLE, yf_negative,  0, 0, 0, 0, 0},
        {"abs",     0, OT_SINGLE, yf_abs,       0, 0, 0, 0, 0},
        {"floor",   0, OT_SINGLE, yf_floor,     0, 0, 0, 0, 0},
        {"ceil",    0, OT_SINGLE, yf_ceil,      0, 0, 0, 0, 0},
        {"sign",    0, OT_SINGLE, yf_sign,      0, 0, 0, 0, 0},
        {"sqrt",    0, OT_SINGLE, yf_sqrt,      0, 0, 0, 0, 0},
        {"exp",     0, OT_SINGLE, yf_exp,       0, 0, 0, 0, 0},
        {"log",     0, OT_SINGLE, yf_log,       0, 0, 0, 0, 0},
        {"log2",    0, OT_SINGLE, yf_log2,      0, 0, 0, 0, 0},
        {"log10",   0, OT_SINGLE, yf_log10,     0, 0, 0, 0, 0},
        {"sin",     0, OT_SINGLE, yf_sin,       0, 0, 0, 0, 0},
        {"cos",     0, OT_SINGLE, yf_cos,       0, 0, 0, 0, 0},
        {"tan",     0, OT_SINGLE, yf_tan,       0, 0, 0, 0, 0},
        {"asin",    0, OT_SINGLE, yf_asin,      0, 0, 0, 0, 0},
        {"acos",    0, OT_SINGLE, yf_acos,      0, 0, 0, 0, 0},
        {"atan",    0, OT_SINGLE, yf_atan,      0, 0, 0, 0, 0},
        {"rand",    0, OT_SINGLE, yf_rand,      0, 0, 0, 0, 0},
    };
    memcpy(g_arrayFunctionOperators + g_countFunctionOperators, singleOperators, sizeof(singleOperators));
    g_countFunctionOperators += SINGLE_OPERATORS_COUNT;

    const Operator twinOperators[TWIN_OPERATORS_COUNT] =
    {
        {"add",     0, OT_TWIN, 0, yf_add,      0, 0, 0, 0},
        {"sub",     0, OT_TWIN, 0, yf_sub,      0, 0, 0, 0},
        {"multiply",0, OT_TWIN, 0, yf_multiply, 0, 0, 0, 0},
        {"divide",  0, OT_TWIN, 0, yf_divide,   0, 0, 0, 0},
        {"max",     0, OT_TWIN, 0, yf_max,      0, 0, 0, 0},
        {"min",     0, OT_TWIN, 0, yf_min,      0, 0, 0, 0},
        {"mod",     0, OT_TWIN, 0, yf_mod,      0, 0, 0, 0},
        {"pow",     0, OT_TWIN, 0, yf_pow,      0, 0, 0, 0},
        {"atan2",   0, OT_TWIN, 0, yf_atan2,    0, 0, 0, 0},
        {"rand2",   0, OT_TWIN, 0, yf_rand2,    0, 0, 0, 0},
    };
    memcpy(g_arrayFunctionOperators + g_countFunctionOperators, twinOperators, sizeof(twinOperators));
    g_countFunctionOperators += TWIN_OPERATORS_COUNT;

    const Operator threeOperators[THREE_OPERATORS_COUNT] =
    {
        {"lerp",    0, OT_THREE, 0, 0, yf_lerp,     0, 0, 0},
        {"clamp",   0, OT_THREE, 0, 0, yf_clamp,    0, 0, 0},
        {"limit",   0, OT_THREE, 0, 0, yf_limit,    0, 0, 0},
        {"gray",    0, OT_THREE, 0, 0, yf_gray,     0, 0, 0},
        {"add3",    0, OT_THREE, 0, 0, yf_add3,     0, 0, 0},
        {"min3",    0, OT_THREE, 0, 0, yf_min3,     0, 0, 0},
        {"max3",    0, OT_THREE, 0, 0, yf_max3,     0, 0, 0},
        {"average3",0, OT_THREE, 0, 0, yf_average3, 0, 0, 0},
    };
    memcpy(g_arrayFunctionOperators + g_countFunctionOperators, threeOperators, sizeof(threeOperators));
    g_countFunctionOperators += THREE_OPERATORS_COUNT;

    const Operator fourOperators[FOUR_OPERATORS_COUNT] = 
    {
        {"add4",    0, OT_FOUR, 0, 0, 0, yf_add4,     0, 0},
        {"min4",    0, OT_FOUR, 0, 0, 0, yf_min4,     0, 0},
        {"max4",    0, OT_FOUR, 0, 0, 0, yf_max4,     0, 0},
        {"average4",0, OT_FOUR, 0, 0, 0, yf_average4, 0, 0},
    };
    memcpy(g_arrayFunctionOperators + g_countFunctionOperators, fourOperators, sizeof(fourOperators));
    g_countFunctionOperators += FOUR_OPERATORS_COUNT;

    const Operator valueNodeOperators[VALUENODE_TO_FLOAT_OPERATORS_COUNT] = 
    {
        {"array_add",0, OT_VALUENODE_TO_FLOAT, 0, 0, 0, 0, yf_add_array, 0},
        {"array_min",0, OT_VALUENODE_TO_FLOAT, 0, 0, 0, 0, yf_min_array, 0},
        {"array_max",0, OT_VALUENODE_TO_FLOAT, 0, 0, 0, 0, yf_max_array, 0},
        {"array_ave",0, OT_VALUENODE_TO_FLOAT, 0, 0, 0, 0, yf_average_array, 0},
    };
    memcpy(g_arrayFunctionOperators + g_countFunctionOperators, valueNodeOperators, sizeof(valueNodeOperators));
    g_countFunctionOperators += VALUENODE_TO_FLOAT_OPERATORS_COUNT;

    const Operator valueNode2Operators[VALUENODE_CONVERT_OPERATORS_COUNT] = 
    {
        {"array_move_right",0, OT_VALUENODE_TO_FLOAT, 0, 0, 0, 0, 0, yf_array_move_right},
        {"array_move_left", 0, OT_VALUENODE_TO_FLOAT, 0, 0, 0, 0, 0, yf_array_move_left},
        {"array_cumulate",  0, OT_VALUENODE_TO_FLOAT, 0, 0, 0, 0, 0, yf_array_cumulate},
    };
    memcpy(g_arrayFunctionOperators + g_countFunctionOperators, valueNode2Operators, sizeof(valueNode2Operators));
    g_countFunctionOperators += VALUENODE_CONVERT_OPERATORS_COUNT;

    for (unsigned int i = 0; i < g_countFunctionOperators; i++)
    {
        g_arrayFunctionOperators[i].hash = get_string_hash(g_arrayFunctionOperators[i].szName);
    }
}

unsigned int GetIndex(const char* szName, unsigned int& hash)
{
    if (!szName || !szName[0])
    {
        return -1;
    }

    size_t len = strlen(szName);
    if (len >= OPERATOR_NAME_LENGTH)
    {
        return -1;
    }

    hash = get_string_hash(szName);

    unsigned int i = 0;
    for (i = 0; i < g_countFunctionOperators; i++)
    {
        if (g_arrayFunctionOperators[i].hash == hash &&
            strcmp(g_arrayFunctionOperators[i].szName, szName) == 0)
        {
            break;
        }
    }

    return i;
}

// ӻ㺯
bool SetSingleOperator(const char* szName, SINGLE_OPERATOR_FUNC func)
{
    unsigned int hash;
    unsigned int index = GetIndex(szName, hash);
    if (index >= MAX_FUNCTIONS_COUNT)
    {
        return false;
    }

    Operator& op = g_arrayFunctionOperators[index];
    strcpy(op.szName, szName);
    op.hash = hash;
    op.type = OT_SINGLE;

    op.single_func = func;
    op.twin_func = 0;
    op.three_func = 0;
    op.four_func = 0;
    op.value_node_to_float_func = 0;
    op.value_node_convert_func = 0;

    if (index == g_countFunctionOperators)
    {
        g_countFunctionOperators++;
    }

    return true;
}

bool SetTwinOperator(const char* szName, TWIN_OPERATOR_FUNC func)
{
    unsigned int hash;
    unsigned int index = GetIndex(szName, hash);
    if (index >= MAX_FUNCTIONS_COUNT)
    {
        return false;
    }

    Operator& op = g_arrayFunctionOperators[index];
    strcpy(op.szName, szName);
    op.hash = hash;
    op.type = OT_TWIN;

    op.single_func = 0;
    op.twin_func = func;
    op.three_func = 0;
    op.four_func = 0;
    op.value_node_to_float_func = 0;
    op.value_node_convert_func = 0;

    if (index == g_countFunctionOperators)
    {
        g_countFunctionOperators++;
    }

    return true;
}

bool SetThreeOperator(const char* szName, THREE_OPERATOR_FUNC func)
{
    unsigned int hash;
    unsigned int index = GetIndex(szName, hash);
    if (index >= MAX_FUNCTIONS_COUNT)
    {
        return false;
    }

    Operator& op = g_arrayFunctionOperators[index];
    strcpy(op.szName, szName);
    op.hash = hash;
    op.type = OT_THREE;

    op.single_func = 0;
    op.twin_func = 0;
    op.three_func = func;
    op.four_func = 0;
    op.value_node_to_float_func = 0;
    op.value_node_convert_func = 0;

    if (index == g_countFunctionOperators)
    {
        g_countFunctionOperators++;
    }

    return true;
}

bool SetFourOperator(const char* szName, FOUR_OPERATOR_FUNC func)
{
    unsigned int hash;
    unsigned int index = GetIndex(szName, hash);
    if (index >= MAX_FUNCTIONS_COUNT)
    {
        return false;
    }

    Operator& op = g_arrayFunctionOperators[index];
    strcpy(op.szName, szName);
    op.hash = hash;
    op.type = OT_FOUR;

    op.single_func = 0;
    op.twin_func = 0;
    op.three_func = 0;
    op.four_func = func;
    op.value_node_to_float_func = 0;
    op.value_node_convert_func = 0;

    if (index == g_countFunctionOperators)
    {
        g_countFunctionOperators++;
    }

    return true;
}

bool SetValueNodeToFloatOperator(const char* szName, VALUENODE_TO_FLOAT_OPERATOR_FUNC func)
{
    unsigned int hash;
    unsigned int index = GetIndex(szName, hash);
    if (index >= MAX_FUNCTIONS_COUNT)
    {
        return false;
    }

    Operator& op = g_arrayFunctionOperators[index];
    strcpy(op.szName, szName);
    op.hash = hash;
    op.type = OT_VALUENODE_TO_FLOAT;

    op.single_func = 0;
    op.twin_func = 0;
    op.three_func = 0;
    op.four_func = 0;
    op.value_node_to_float_func = func;
    op.value_node_convert_func = 0;

    if (index == g_countFunctionOperators)
    {
        g_countFunctionOperators++;
    }

    return true;
}

bool SetValueNodeConvertOperator(const char* szName, VALUENODE_CONVERT_OPERATOR_FUNC func)
{
    unsigned int hash;
    unsigned int index = GetIndex(szName, hash);
    if (index >= MAX_FUNCTIONS_COUNT)
    {
        return false;
    }

    Operator& op = g_arrayFunctionOperators[index];
    strcpy(op.szName, szName);
    op.hash = hash;
    op.type = OT_VALUENODE_CONVERT;

    op.single_func = 0;
    op.twin_func = 0;
    op.three_func = 0;
    op.four_func = 0;
    op.value_node_to_float_func = 0;
    op.value_node_convert_func = func;

    if (index == g_countFunctionOperators)
    {
        g_countFunctionOperators++;
    }

    return true;
}

// ȡ㺯
const Operator* GetFunctionOperator(const char* szName)
{
    if (g_countFunctionOperators == 0)
    {
        InitFunctionOperators();
    }

    unsigned int hash;
    unsigned int index = GetIndex(szName, hash);
    if (index >= g_countFunctionOperators)
    {
        return NULL;
    }

    return g_arrayFunctionOperators + index;
}

// --------------------------------------------------------------------------------------

};