
/****************************************************************

  File name   :  value_node.h
  Author      :  Ҷ
  Version     :  1.0a
  Create Date :  2014/06/18 
  Description :  ֵڵ 
  Others      :  

*****************************************************************/

#ifndef __value_node_H__
#define __value_node_H__

// --------------------------------------------------------------------------------------

#include <cstdlib>
#include <cstring>
#include <cmath>
#include <cfloat>
#include <cstdio>

// --------------------------------------------------------------------------------------

namespace Why
{

struct ValueNode
{
    float _value;               // ֵڵ
    float* _arrayValues;        // ֵڵ
    unsigned int _valueCount;   // 鳤

    ValueNode()
    {
        _value = 0.0f;
        _arrayValues = NULL;
        _valueCount = 0;
    }

    ~ValueNode()
    {
        if (_arrayValues)
        {
            free(_arrayValues);
        }
    }

    void Clear()
    {
        _value = 0.0f;
        if (_arrayValues)
        {
            free(_arrayValues);
            _arrayValues = NULL;
        }
        _valueCount = 0;
    }

    void Resize(unsigned int count)
    {
        if (_valueCount == count)
        {
            return;
        }

        float* newArrayValues = NULL;
        if (count > 0)
        {
            newArrayValues = (float*)malloc(sizeof(float)*count);
            if (!_arrayValues)
            {
                if (_value > -FLT_EPSILON && _value < FLT_EPSILON)
                {
                    memset(newArrayValues, 0, sizeof(float)*count);
                }
                else
                {
                    for (unsigned int i = 0; i < count; i++)
                    {
                        newArrayValues[i] = _value;
                    }
                    _value = 0.0f;
                }
            }
            else
            {
                if (count <= _valueCount)
                {
                    memcpy(newArrayValues, _arrayValues, sizeof(float)*count);
                }
                else
                {
                    memcpy(newArrayValues, _arrayValues, sizeof(float)*_valueCount);
                    memset(newArrayValues + _valueCount, 0, sizeof(float)*(count - _valueCount));
                }
            }
        }

        if (_arrayValues)
        {
            free(_arrayValues);
        }
        _arrayValues = newArrayValues;

        _valueCount = count;
    }

    void SetValue(float v)
    {
        Clear();
        _value = v;
    }

    void SetValue(const float* ptr, unsigned int count)
    {
        Resize(count);
        if (count)
        {
            memcpy(_arrayValues, ptr, sizeof(float)*count);
        }
    }

    // Բֵ
    bool LinerInterpolate(float vStart, float vEnd)
    {
        if (_valueCount < 2)
        {
            return false;
        }
        float step = (vEnd - vStart) / (_valueCount - 1);

        for (unsigned int i = 0; i < _valueCount; i++)
        {
            _arrayValues[i] = vStart + step*i;
        }

        return true;
    }

    float GetValueByIndex(unsigned int index) const
    {
        if (!_arrayValues)
        {
            return _value;
        }
        else if (index >= _valueCount)
        {
            return 0.0f;
        }
        else
        {
            return _arrayValues[index];
        }
    }

    bool SetValueByIndex(unsigned int index, float v) const
    {
        if (!_arrayValues || index >= _valueCount)
        {
            return false;
        }

        _arrayValues[index] = v;

        return true;
    }

    float GetMinValue() const
    {
        if (!_arrayValues)
        {
            return _value;
        }

        float m = _arrayValues[0];

        for (unsigned int i = 1; i < _valueCount; i++)
        {
            if (m > _arrayValues[i])
            {
                m = _arrayValues[i];
            }
        }

        return m;
    }

    float GetMaxValue() const
    {
        if (!_arrayValues)
        {
            return _value;
        }

        float m = _arrayValues[0];

        for (unsigned int i = 1; i < _valueCount; i++)
        {
            if (m < _arrayValues[i])
            {
                m = _arrayValues[i];
            }
        }

        return m;
    }

    void Copy(const ValueNode& another)
    {
        if (another._valueCount)
        {
            Resize(another._valueCount);
            memcpy(_arrayValues, another._arrayValues, sizeof(float)*_valueCount);
        }
        else
        {
            SetValue(another._value);
        }
    }

    void SingleDuplicate(unsigned int n)
    {
        if (!_valueCount)
        {
            return;
        }

        float* buf = (float*)malloc(sizeof(float)*_valueCount*n);

        for (unsigned int j = 0; j < _valueCount; j++)
        {
            for (unsigned int i = 0; i < n; i++)
            {
                buf[j*n + i] = _arrayValues[j];
            }
        }

        free(_arrayValues);
        _arrayValues = buf;
        _valueCount *= n;
    }

    void WholeDuplicate(unsigned int n)
    {
        if (!_valueCount)
        {
            return;
        }

        float* buf = (float*)malloc(sizeof(float)*_valueCount*n);
        for (unsigned int i = 0; i < n; i++)
        {
            memcpy(buf + i*_valueCount, _arrayValues, sizeof(float)*_valueCount);
        }

        free(_arrayValues);
        _arrayValues = buf;
        _valueCount *= n;
    }
};

};

// --------------------------------------------------------------------------------------

#endif // __value_node_H__
