
/****************************************************************

  File name   :  value_node_operate.cpp
  Author      :  Ҷ
  Version     :  1.0a
  Create Date :  2014/06/18 
  Description :   
  Others      :  

*****************************************************************/

// --------------------------------------------------------------------------------------

#include "value_node_operate.h"

// --------------------------------------------------------------------------------------

namespace Why
{

// --------------------------------------------------------------------------------------

void ValueNodeSingleOperate(
    SINGLE_OPERATOR_FUNC single_func, 
    ValueNode* pOut, 
    const ValueNode* p)
{
    if (p->_valueCount)
    {
        pOut->SetValue(p->_arrayValues, p->_valueCount);
        for (unsigned int i = 0; i < p->_valueCount; i++)
        {
            pOut->_arrayValues[i] = single_func(pOut->_arrayValues[i]);
        }
    }
    else
    {
        float v = single_func(p->_value);
        pOut->SetValue(v);
    }
}

void ValueNodeTwinOperate(
    TWIN_OPERATOR_FUNC twin_func, 
    ValueNode* pOut, 
    const ValueNode* p1, 
    const ValueNode* p2)
{
    unsigned int count = p1->_valueCount > p2->_valueCount ? p1->_valueCount : p2->_valueCount;

    if (count == 0)
    {
        float v = twin_func(p1->_value, p2->_value);
        pOut->SetValue(v);
    }
    else
    {
        pOut->Resize(count);
        float v1;
        float v2;
        float v;
        for (unsigned int i = 0; i < count; i++)
        {
            v1 = p1->GetValueByIndex(i);
            v2 = p2->GetValueByIndex(i);
            v = twin_func(v1, v2);
            pOut->SetValueByIndex(i, v);
        }
    }
}

void ValueNodeThreeOperate(
    THREE_OPERATOR_FUNC three_func, 
    ValueNode* pOut, 
    const ValueNode* p1, 
    const ValueNode* p2, 
    const ValueNode* p3)
{
    unsigned int count = p1->_valueCount > p2->_valueCount ? p1->_valueCount : p2->_valueCount;
    if (count < p3->_valueCount)
    {
        count = p3->_valueCount;
    }

    if (0 == count)
    {
        float v = three_func(p1->_value, p2->_value, p3->_value);
        pOut->SetValue(v);
    }
    else
    {
        pOut->Resize(count);
        float v1;
        float v2;
        float v3;
        float v;
        for (unsigned int i = 0; i < count; i++)
        {
            v1 = p1->GetValueByIndex(i);
            v2 = p2->GetValueByIndex(i);
            v3 = p3->GetValueByIndex(i);
            v = three_func(v1, v2, v3);
            pOut->SetValueByIndex(i, v);
        }
    }
}

void ValueNodeFourOperate(
    FOUR_OPERATOR_FUNC four_func,
    ValueNode* pOut, 
    const ValueNode* p1, 
    const ValueNode* p2, 
    const ValueNode* p3,
    const ValueNode* p4)
{
    unsigned int count = p1->_valueCount > p2->_valueCount ? p1->_valueCount : p2->_valueCount;
    if (count < p3->_valueCount)
    {
        count = p3->_valueCount;
    }
    if (count < p4->_valueCount)
    {
        count = p4->_valueCount;
    }

    if (0 == count)
    {
        float v = four_func(p1->_value, p2->_value, p3->_value, p4->_value);
        pOut->SetValue(v);
    }
    else
    {
        pOut->Resize(count);
        float v1;
        float v2;
        float v3;
        float v4;
        float v;
        for (unsigned int i = 0; i < count; i++)
        {
            v1 = p1->GetValueByIndex(i);
            v2 = p2->GetValueByIndex(i);
            v3 = p3->GetValueByIndex(i);
            v4 = p4->GetValueByIndex(i);
            v = four_func(v1, v2, v3, v4);
            pOut->SetValueByIndex(i, v);
        }
    }
}

// --------------------------------------------------------------------------------------

};