
/****************************************************************

  File name   :  MathExpressionCommon.cpp
  Author      :  Ҷ
  Version     :  1.0a
  Create Date :  2014/01/08 
  Description :   
  Others      :  

*****************************************************************/

// --------------------------------------------------------------------------------------

#include <cstdlib>
#include <cmath>
#include <cfloat>
#include <cstring>
#include <cstdio>
#include <assert.h>

#include "MathExpressionCommon.h"

// --------------------------------------------------------------------------------------

#pragma   warning(disable: 4996)

// --------------------------------------------------------------------------------------

#define REAL_PI                     3.14159265358979f               // Բ
#define REAL_E                      2.71828182845905f               // e

// --------------------------------------------------------------------------------------

// ôдĸ

const ConstValue g_constValues[CONST_VALUES_COUNT] = 
{
    {"PI",      REAL_PI},
    {"E",       REAL_E},
};            

// --------------------------------------------------------------------------------------

const Operator g_standardSingleOperators[STANDARD_SINGLE_OPERATORS_COUNT] =
{
    {"+",       1, OT_STANDARD_SINGLE, yf_positive, 0, 0, 0},
    {"-",       1, OT_STANDARD_SINGLE, yf_negative, 0, 0, 0},
};

const Operator g_standardTwinOperators[STANDARD_TWIN_OPERATORS_COUNT] =
{
    {"+",       1, OT_STANDARD_TWIN, 0, yf_add, 0, 0},
    {"-",       1, OT_STANDARD_TWIN, 0, yf_sub, 0, 0},
    {"*",       1, OT_STANDARD_TWIN, 0, yf_multiply, 0, 0},
    {"/",       1, OT_STANDARD_TWIN, 0, yf_divide, 0, 0},
    {"%",       1, OT_STANDARD_TWIN, 0, yf_mod, 0, 0},
    {"^",       1, OT_STANDARD_TWIN, 0, yf_pow, 0, 0},
};

// Сдĸ 

const Operator g_singleOperators[SINGLE_OPERATORS_COUNT] =
{
    {"positive",8, OT_SINGLE, yf_positive, 0, 0, 0},
    {"negative",8, OT_SINGLE, yf_negative, 0, 0, 0},
    {"abs",     3, OT_SINGLE, yf_abs, 0, 0, 0},
    {"floor",   5, OT_SINGLE, yf_floor, 0, 0, 0},
    {"ceil",    4, OT_SINGLE, yf_ceil, 0, 0, 0},
    {"sign",    4, OT_SINGLE, yf_sign, 0, 0, 0},
    {"sqrt",    4, OT_SINGLE, yf_sqrt, 0, 0, 0},
    {"exp",     3, OT_SINGLE, yf_exp, 0, 0, 0},
    {"log",     3, OT_SINGLE, yf_log, 0, 0, 0},
    {"log2",    4, OT_SINGLE, yf_log2, 0, 0, 0},
    {"log10",   5, OT_SINGLE, yf_log10, 0, 0, 0},
    {"sin",     3, OT_SINGLE, yf_sin, 0, 0, 0},
    {"cos",     3, OT_SINGLE, yf_cos, 0, 0, 0},
    {"tan",     3, OT_SINGLE, yf_tan, 0, 0, 0},
    {"asin",    4, OT_SINGLE, yf_asin, 0, 0, 0},
    {"acos",    4, OT_SINGLE, yf_acos, 0, 0, 0},
    {"atan",    4, OT_SINGLE, yf_atan, 0, 0, 0},
    {"rand",    4, OT_SINGLE, yf_rand, 0, 0, 0},
};

const Operator g_twinOperators[TWIN_OPERATORS_COUNT] =
{
    {"add",     3, OT_TWIN, 0, yf_add, 0, 0},
    {"sub",     3, OT_TWIN, 0, yf_sub, 0, 0},
    {"multiply",8, OT_TWIN, 0, yf_multiply, 0, 0},
    {"divide",  6, OT_TWIN, 0, yf_divide, 0, 0},
    {"max",     3, OT_TWIN, 0, yf_max, 0, 0},
    {"min",     3, OT_TWIN, 0, yf_min, 0, 0},
    {"mod",     3, OT_TWIN, 0, yf_mod, 0, 0},
    {"pow",     3, OT_TWIN, 0, yf_pow, 0, 0},
    {"atan2",   5, OT_TWIN, 0, yf_atan2, 0, 0},
    {"rand2",   5, OT_TWIN, 0, yf_rand2, 0, 0},
};

const Operator g_threeOperators[THREE_OPERATORS_COUNT] =
{
    {"lerp",    4, OT_TWIN, 0, 0, yf_lerp, 0},
    {"clamp",   5, OT_TWIN, 0, 0, yf_clamp, 0},
    {"gray",    4, OT_TWIN, 0, 0, yf_gray, 0},
    {"add3",    4, OT_TWIN, 0, 0, yf_add3, 0},
    {"average3",8, OT_TWIN, 0, 0, yf_average3, 0},
};

const Operator g_fourOperators[FOUR_OPERATORS_COUNT] = 
{
    {"add4",    4, OT_TWIN, 0, 0, 0, yf_add4},
    {"average4",8, OT_TWIN, 0, 0, 0, yf_average4},
};

const Operator g_leftBracketOperator  = {"([{", 3, OT_LEFT_BRACKET, 0, 0, 0, 0};
const Operator g_rightBracketOperator = {")]}", 3, OT_RIGHT_BRACKET, 0, 0, 0, 0};

// --------------------------------------------------------------------------------------

// 

float     yf_positive(float a)
{
    return a;
}

float     yf_negative(float a)
{
    return -a;
}

float     yf_add(float a, float b)
{
    return a + b;
}

float     yf_sub(float a, float b)
{
    return a - b;
}

float     yf_multiply(float a, float b)
{
    return a * b;
}

float     yf_divide(float a, float b)
{
    if (b < FLT_EPSILON)
    {
        return a > 0 ? FLT_MAX : -FLT_MAX;
    }
    else
    {
        return a / b;
    }
}

float     yf_abs(float a)
{
    return ::fabsf(a);
}

float     yf_floor(float a)
{
    return ::floorf(a);
}

float     yf_ceil(float a)
{
    return ::ceilf(a);
}

float     yf_sign(float a)
{
    return (a >= 0.0f) ? 1.0f : -1.0f;
}

float     yf_sqrt(float a)
{
    if (a < 0.0f)
    {
        return 0.0f;
    }
    else
    {
        return ::sqrtf(a);
    }
}

float     yf_exp(float a)
{
    return ::expf(a);
}

float     yf_log(float a)
{
    return ::logf(a);
}

float     yf_log2(float a)
{
    const float ln2 = (float)0.693147180559945309417;
    return ::logf(a) / ln2;
}

float     yf_log10(float a)
{
    return ::log10f(a);
}

float     yf_sin(float a)
{
    return ::sinf(a);
}

float     yf_cos(float a)
{
    return ::cosf(a);
}

float     yf_tan(float a)
{
    return ::tanf(a);
}

float     yf_asin(float f)
{
    if(f >= 1.0f)   return REAL_PI*0.5f;
    if(f <= -1.0f)  return -REAL_PI*0.5f;
    return ::asinf(f);
}

float     yf_acos(float f)
{
    if(f >= 1.0f)   return 0.0f;
    if(f <= -1.0f)  return REAL_PI;
    return ::acosf(f);
}

float     yf_atan(float a)
{
    return ::atanf(a);
}

float     yf_rand(float a)
{
    return a * (float)(::rand() / ((float)RAND_MAX + 1));
}

// --------------------------------------------------------------------------------------

float     yf_max(float a, float b)
{
    return (a < b) ? b : a;
}

float     yf_min(float a, float b)
{
    return (a < b) ? a : b;
}

float     yf_mod(float a, float b)
{
    return ::fmodf(a, b);
}

float     yf_pow(float a, float b)
{
    return ::powf(a, b);
}

float     yf_atan2(float x, float y)
{
    return ::atan2f(x, y);
}

float     yf_rand2(float a, float b)
{
    const float r = (float)(::rand() / ((float)RAND_MAX + 1));
    return r*(b-a) + a;
}

// --------------------------------------------------------------------------------------

float     yf_lerp(float a, float b, float r)
{
    return a + (b - a) * r;
}

float     yf_clamp(float v, float hi, float low)
{
    if (v > hi) 
        return hi;
    else if (v < low) 
        return low;
    else
        return v;
}

float     yf_gray(float r, float g, float b)
{
    return r*0.299f + g*0.587f + b*0.114f;
}

float     yf_add3(float a, float b, float c)
{
    return a+b+c;
}

float     yf_average3(float a, float b, float c)
{
    return (a+b+c)/3.0f;
}

// --------------------------------------------------------------------------------------

float     yf_add4(float a, float b, float c, float d)
{
    return a+b+c+d;
}

float     yf_average4(float a, float b, float c, float d)
{
    return (a+b+c+d)*0.25f;
}

// --------------------------------------------------------------------------------------

// ǷΪСдĸ
inline bool is_small_letter(char c)
{
    return c >= 'a' && c <= 'z';
}

// ǷΪдĸ
inline bool is_big_letter(char c)
{
    return c >= 'A' && c <= 'Z';
}

// ǷΪ
inline bool is_number(char c)
{
    return c >= '0' && c <= '9';
}

// ǷΪָ
inline bool is_separator(char c)
{
    return (' ' == c || 0 == c || ',' == c || 
        '(' == c || ')' == c || '[' == c || ']' == c || '{' == c || '}' == c || '<' == c || '>' == c || 
        '+' == c || '-' == c || '*' == c || '/' == c || '%' == c || '^' == c);
}

// ַת,֧2,8,10,16ֽ
/*
    0XͷΪ16, XABCDEFСд, 0xffff   
    0YͷΪ2, YСд, 0y10101010000
    0ͷΪ8, 07523, ע"08"֮ʧ
    ĬΪ10
*/
bool     string_to_int(const char* szSrc, int& rst)
{
    rst = 0;

    char* str = (char*)szSrc;
    if(str[0] == '0')
    {
        if (str[1] == 'x' || str[1] == 'X')
        {
            // 16
            str += 2;
            int nDecNum;
            int len = (int)strlen(str);
            int n = 1;
            for(int i = len - 1; i >= 0; i--)
            {
                switch(str[i])
                {
                case 'a':
                case 'A':
                    nDecNum = 10;
                    break;
                case 'b':
                case 'B':
                    nDecNum = 11;
                    break;
                case 'c':
                case 'C':
                    nDecNum = 12;
                    break;
                case 'd':
                case 'D':
                    nDecNum = 13;
                    break;
                case 'e':
                case 'E':
                    nDecNum = 14;
                    break;
                case 'f':
                case 'F':
                    nDecNum = 15;
                    break;
                case '0':
                case '1':
                case '2':
                case '3':
                case '4':
                case '5':
                case '6':
                case '7':
                case '8':
                case '9':
                    nDecNum = str[i] - '0';
                    break;
                default:
                    return false;
                }
                rst += nDecNum * n;
                n *= 16;
            }
            return true;
        }
        else if (str[1] == 'y' || str[1] == 'Y')
        {
            // 2
            str += 2;
            int len = (int)strlen(str);
            int n = 1;
            for(int i = len - 1; i >= 0; i--)
            {
                if (str[i] == '1')
                {
                    rst += n;
                }
                else if (str[i] != '0')
                {
                    return false;
                }
                n *= 2;
            }
            return true;
        }
        else
        {
            // 8
            str += 1;
            int nDecNum;
            int len = (int)strlen(str);
            int n = 1;
            for(int i = len - 1; i >= 0; i--)
            {
                nDecNum = str[i] - '0';
                if (nDecNum > 7 || nDecNum < 0)
                {
                    return false;
                }
                rst += nDecNum * n;
                n *= 8;
            }
            return true;
        }
    }
    else
    {
        int len = (int)strlen(str);
        for (int i = 0; i < len; i++)
        {
            if (!is_number(str[i]))
            {
                return false;
            }
        }
        // 10
        rst = ::atoi(str);
        return true;
    }
}

// תΪһƵַ
void int_to_string(int value, char* szDest, int type = 10)
{
    bool negative;
    if (value < 0)
    {
        negative = true;
        value = -value;
    }

    int index = 0;
    int nDecNum;
    char szTmp[128];
    memset(szTmp, 0, 128);
    while (value && index < 128)
    {
        nDecNum = value % type;
        value /= type;
        if (nDecNum < 10)
        {
            szTmp[index] = '0' + nDecNum;
        }
        else
        {
            szTmp[index] = 'A' + nDecNum - 10;
        }
        index++;
    }

    int pos = 0;
    if (negative)
    {
        szDest[pos] = '-';
        pos++;
    }
    for (int i = index - 1; i >= 0; i--)
    {
        szDest[pos] = szTmp[i];
        pos++;
    }
    szDest[pos] = 0;
}

// --------------------------------------------------------------------------------------

unsigned int IsLeftBracketLetter(char c)
{
    for (unsigned int i = 0; i < g_leftBracketOperator.nameLen; i++)
    {
        if (c == g_leftBracketOperator.szName[i])
        {
            return i;
        }
    }

    return -1;
}

unsigned int IsRightBracketLetter(char c)
{
    for (unsigned int i = 0; i < g_rightBracketOperator.nameLen; i++)
    {
        if (c == g_rightBracketOperator.szName[i])
        {
            return i;
        }
    }

    return -1;
}

const Operator* GetLeftBracketOperator(const char* szExpression, unsigned int& pos, unsigned int& index)
{
    while (szExpression[pos] == ' ')
    {
        pos++;
    }

    char cOperator = szExpression[pos];

    index = IsLeftBracketLetter(cOperator);
    if (index != -1)
    {
        pos++;
        return &g_leftBracketOperator;
    }

    return NULL;
}

const Operator* GetRightBracketOperator(const char* szExpression, unsigned int& pos, unsigned int& index)
{
    while (szExpression[pos] == ' ')
    {
        pos++;
    }

    char cOperator = szExpression[pos];

    index = IsRightBracketLetter(cOperator);
    if (index != -1)
    {
        pos++;
        return &g_rightBracketOperator;
    }

    return NULL;
}

const Operator* GetStandardSingleOperator(const char* szExpression, unsigned int& pos)
{
    while (szExpression[pos] == ' ')
    {
        pos++;
    }

    char cOperator = szExpression[pos];
    for (unsigned int i = 0; i < STANDARD_SINGLE_OPERATORS_COUNT; i++)
    {
        if (g_standardSingleOperators[i].szName[0] == cOperator)
        {
            pos++;
            return g_standardSingleOperators + i;
        }
    }

    return NULL;
}

const Operator* GetStandardTwinOperator(const char* szExpression, unsigned int& pos)
{
    while (szExpression[pos] == ' ')
    {
        pos++;
    }

    char cOperator = szExpression[pos];
    for (unsigned int i = 0; i < STANDARD_TWIN_OPERATORS_COUNT; i++)
    {
        if (g_standardTwinOperators[i].szName[0] == cOperator)
        {
            pos++;
            return g_standardTwinOperators + i;
        }
    }

    return NULL;
}

const Operator* GetSingleOperator(const char* szExpression, unsigned int& pos)
{
    while (szExpression[pos] == ' ')
    {
        pos++;
    }

    if (!is_small_letter(szExpression[pos]))
    {
        return NULL;
    }

    char szName[NAME_MAX_LENGTH];
    memset(szName, 0, NAME_MAX_LENGTH);
    unsigned int nameLen = 0;

    for (nameLen = 0; nameLen < NAME_MAX_LENGTH - 1; nameLen++)
    {
        if (!is_small_letter(szExpression[pos + nameLen]) && !is_number(szExpression[pos + nameLen]) )
        {
            break;
        }
        szName[nameLen] = szExpression[pos + nameLen];
    }

    for (unsigned int i = 0; i < SINGLE_OPERATORS_COUNT; i++)
    {
        if (nameLen == g_singleOperators[i].nameLen &&
            strcmp(g_singleOperators[i].szName, szName) == 0)
        {
            pos += nameLen;
            return g_singleOperators + i;
        }
    }

    return NULL;
}

const Operator* GetTwinOperator(const char* szExpression, unsigned int& pos)
{
    while (szExpression[pos] == ' ')
    {
        pos++;
    }

    if (!is_small_letter(szExpression[pos]))
    {
        return NULL;
    }

    char szName[NAME_MAX_LENGTH];
    memset(szName, 0, NAME_MAX_LENGTH);
    unsigned int nameLen = 0;

    for (nameLen = 0; nameLen < NAME_MAX_LENGTH - 1; nameLen++)
    {
        if (!is_small_letter(szExpression[pos + nameLen]) && !is_number(szExpression[pos + nameLen]) )
        {
            break;
        }
        szName[nameLen] = szExpression[pos + nameLen];
    }

    for (unsigned int i = 0; i < TWIN_OPERATORS_COUNT; i++)
    {
        if (nameLen == g_twinOperators[i].nameLen &&
            strcmp(g_twinOperators[i].szName, szName) == 0)
        {
            pos += nameLen;
            return g_twinOperators + i;
        }
    }

    return NULL;
}

const Operator* GetThreeOperator(const char* szExpression, unsigned int& pos)
{
    while (szExpression[pos] == ' ')
    {
        pos++;
    }

    if (!is_small_letter(szExpression[pos]))
    {
        return NULL;
    }

    char szName[NAME_MAX_LENGTH];
    memset(szName, 0, NAME_MAX_LENGTH);
    unsigned int nameLen = 0;

    for (nameLen = 0; nameLen < NAME_MAX_LENGTH - 1; nameLen++)
    {
        if (!is_small_letter(szExpression[pos + nameLen]) && !is_number(szExpression[pos + nameLen]) )
        {
            break;
        }
        szName[nameLen] = szExpression[pos + nameLen];
    }

    for (unsigned int i = 0; i < THREE_OPERATORS_COUNT; i++)
    {
        if (nameLen == g_threeOperators[i].nameLen &&
            strcmp(g_threeOperators[i].szName, szName) == 0)
        {
            pos += nameLen;
            return g_threeOperators + i;
        }
    }

    return NULL;
}

const Operator* GetFourOperator(const char* szExpression, unsigned int& pos)
{
    while (szExpression[pos] == ' ')
    {
        pos++;
    }

    if (!is_small_letter(szExpression[pos]))
    {
        return NULL;
    }

    char szName[NAME_MAX_LENGTH];
    memset(szName, 0, NAME_MAX_LENGTH);
    unsigned int nameLen = 0;

    for (nameLen = 0; nameLen < NAME_MAX_LENGTH - 1; nameLen++)
    {
        if (!is_small_letter(szExpression[pos + nameLen]) && !is_number(szExpression[pos + nameLen]) )
        {
            break;
        }
        szName[nameLen] = szExpression[pos + nameLen];
    }

    for (unsigned int i = 0; i < FOUR_OPERATORS_COUNT; i++)
    {
        if (nameLen == g_fourOperators[i].nameLen &&
            strcmp(g_fourOperators[i].szName, szName) == 0)
        {
            pos += nameLen;
            return g_fourOperators + i;
        }
    }

    return NULL;
}

// Ҷ
bool GetCommaOperator(const char* szExpression, unsigned int& pos)
{
    while (szExpression[pos] == ' ')
    {
        pos++;
    }

    if (',' == szExpression[pos])
    {
        pos++;
        return true;
    }

    return false;
}

// Ҳ x,y,z
char GetParameter(const char* szExpression, unsigned int& pos)
{
    while (szExpression[pos] == ' ')
    {
        pos++;
    }

    if (!is_separator(szExpression[pos + 1]))
    {
        return 0;
    }

    char c = szExpression[pos];
    if (c == 'x' || c == 'X')
    {
        pos++;
        return 'x';
    }
    else if (c == 'y' || c == 'Y')
    {
        pos++;
        return 'y';
    } 
    else if (c == 'z' || c == 'Z')
    {
        pos++;
        return 'z';
    }
    else
    {
        return 0;
    }
}

bool GetConstValue(float& rst, const char* szExpression, unsigned int& pos)
{
    while (szExpression[pos] == ' ')
    {
        pos++;
    }

    if (!is_big_letter(szExpression[pos]))
    {
        return false;
    }

    char szName[NAME_MAX_LENGTH];
    memset(szName, 0, NAME_MAX_LENGTH);
    unsigned int nameLen = 0;

    for (nameLen = 0; nameLen < NAME_MAX_LENGTH - 1; nameLen++)
    {
        if (!is_big_letter(szExpression[pos + nameLen]) && !is_number(szExpression[pos + nameLen]) )
        {
            break;
        }
        szName[nameLen] = szExpression[pos + nameLen];
    }

    for (unsigned int i = 0; i < CONST_VALUES_COUNT; i++)
    {
        if (strcmp(g_constValues[i].szName, szName) == 0)
        {
            pos += nameLen;
            rst = g_constValues[i].value;
            return true;
        }
    }

    return false;
}

bool GetFloatValue(float& rst, const char* szExpression, unsigned int& pos)
{
    while (szExpression[pos] == ' ')
    {
        pos++;
    }

    if (!is_number(szExpression[pos]))
    {
        return false;
    }

    char szValue[64];
    memset(szValue, 0, 64);
    unsigned int valueLen = 0;

    bool dot = false;

    for (valueLen = 0; valueLen < 63; valueLen++)
    {
        if (szExpression[pos + valueLen] == '.')
        {
            if (dot)
            {
                return false;
            }
            dot = true;
        }
        else if (is_separator(szExpression[pos + valueLen]))
        {
            break;
        }
        else if (!is_number(szExpression[pos + valueLen]))
        {
            return false;
        }

        szValue[valueLen] = szExpression[pos + valueLen];
    }

    if (!dot)
    {
        return false;
    }

    pos += valueLen;
    rst = (float)::atof(szValue);

    return true;
}

bool GetIntegerValue(float& rst, const char* szExpression, unsigned int& pos)
{
    while (szExpression[pos] == ' ')
    {
        pos++;
    }

    if (!is_number(szExpression[pos]))
    {
        return false;
    }

    char szValue[64];
    memset(szValue, 0, 64);
    unsigned int valueLen = 0;

    bool dot = false;

    for (valueLen = 0; valueLen < 63; valueLen++)
    {
        if (is_separator(szExpression[pos + valueLen]))
        {
            break;
        }
        szValue[valueLen] = szExpression[pos + valueLen];
    }

    int tmp;
    if (string_to_int(szValue, tmp))
    {
        pos += valueLen;
        rst = (float)tmp;
        return true;
    }
    else
    {
        return false;
    }
}

// --------------------------------------------------------------------------------------
