
/****************************************************************

  File name   :  MathExpressionParser.cpp
  Author      :  Ҷ
  Version     :  1.0a
  Create Date :  2014/01/08 
  Description :  ѧʽ 
  Others      :  

*****************************************************************/

// --------------------------------------------------------------------------------------

#include <cstring>
#include <cmath>
#include <assert.h>

#include "MathExpressionCommon.h"
#include "MathExpressionParser.h"

// --------------------------------------------------------------------------------------

#pragma   warning(disable: 4996)

// --------------------------------------------------------------------------------------

// ڵӱʽ
bool BracketOperatorExpressionParser(const char* szExpression, unsigned int& pos, float& rst, 
                                     float x, float y, float z, 
                                     unsigned int bracketIndex, bool endComma)
{
    rst = 0.0f;

    int nLeftBracket = 1;
    unsigned int bracketIndices[MAX_BRACKET_SIZE];
    bracketIndices[0] = bracketIndex;

    char szSubExp[SUB_EXPRESSION_MAX_LENGTH];
    memset(szSubExp, 0, SUB_EXPRESSION_MAX_LENGTH);

    unsigned int leftBI;
    unsigned int rightBI;

    unsigned int subExpLen = 0;
    while (szExpression[pos + subExpLen])
    {
        leftBI = IsLeftBracketLetter(szExpression[pos + subExpLen]);
        rightBI = IsRightBracketLetter(szExpression[pos + subExpLen]);

        if (leftBI != -1)
        {
            if (nLeftBracket > MAX_BRACKET_SIZE - 1)
            {
                return false;
            }
            bracketIndices[nLeftBracket] = leftBI;
            nLeftBracket++;
        }
        else if (rightBI != -1)
        {
            nLeftBracket--;
            if (bracketIndices[nLeftBracket] != rightBI)
            {
                return false;
            }
        }
        else if (endComma && szExpression[pos + subExpLen] == ',')
        {
            // Զβ
            if (nLeftBracket == 1)
            {
                nLeftBracket--;
            }
            else
            {
                return false;
            }
        }

        if (nLeftBracket == 0)
        {
            break;
        }
        szSubExp[subExpLen] = szExpression[pos + subExpLen];
        subExpLen++;
    }

    if (nLeftBracket != 0)
    {
        return false;
    }

    pos += subExpLen + 1;

    return MathExpressionParser(szSubExp, rst, x, y, z);
}

// Ŀӱʽ
bool SingleOperatorExpressionParser(const char* szExpression, unsigned int& pos, const Operator* operatorPtr, float& rst, float x, float y, float z)
{
    rst = 0.0f;
    if (!operatorPtr || !operatorPtr->single_func)
    {
        return false;
    }

    unsigned int bracketIndex;
    if (!GetLeftBracketOperator(szExpression, pos, bracketIndex))
    {
        return false;
    }

    if (!BracketOperatorExpressionParser(szExpression, pos, rst, x, y, z, bracketIndex, false))
    {
        return false;
    }

    rst = operatorPtr->single_func(rst);

    return true;
}

// ˫Ŀӱʽ
bool TwinOperatorExpressionParser(const char* szExpression, unsigned int& pos, const Operator* operatorPtr, float& rst, float x, float y, float z)
{
    rst = 0.0f;
    if (!operatorPtr || !operatorPtr->twin_func)
    {
        return false;
    }

    unsigned int bracketIndex;
    if (!GetLeftBracketOperator(szExpression, pos, bracketIndex))
    {
        return false;
    }

    float valueLeft;
    if (!BracketOperatorExpressionParser(szExpression, pos, valueLeft, x, y, z, bracketIndex, true))
    {
        return false;
    }

    float valueRight;
    if (!BracketOperatorExpressionParser(szExpression, pos, valueRight, x, y, z, bracketIndex, false))
    {
        return false;
    }

    rst = operatorPtr->twin_func(valueLeft, valueRight);

    return true;
}

// Ŀӱʽ
bool ThreeOperatorExpressionParser(const char* szExpression, unsigned int& pos, const Operator* operatorPtr, float& rst, float x, float y, float z)
{
    rst = 0.0f;
    if (!operatorPtr || !operatorPtr->three_func)
    {
        return false;
    }

    unsigned int bracketIndex;
    if (!GetLeftBracketOperator(szExpression, pos, bracketIndex))
    {
        return false;
    }

    float value1;
    if (!BracketOperatorExpressionParser(szExpression, pos, value1, x, y, z, bracketIndex, true))
    {
        return false;
    }

    float value2;
    if (!BracketOperatorExpressionParser(szExpression, pos, value2, x, y, z, bracketIndex, true))
    {
        return false;
    }

    float value3;
    if (!BracketOperatorExpressionParser(szExpression, pos, value3, x, y, z, bracketIndex, false))
    {
        return false;
    }

    rst = operatorPtr->three_func(value1, value2, value3);

    return true;
}

// Ŀӱʽ
bool FourOperatorExpressionParser(const char* szExpression, unsigned int& pos, const Operator* operatorPtr, float& rst, float x, float y, float z)
{
    rst = 0.0f;
    if (!operatorPtr || !operatorPtr->four_func)
    {
        return false;
    }

    unsigned int bracketIndex;
    if (!GetLeftBracketOperator(szExpression, pos, bracketIndex))
    {
        return false;
    }

    float value1;
    if (!BracketOperatorExpressionParser(szExpression, pos, value1, x, y, z, bracketIndex, true))
    {
        return false;
    }

    float value2;
    if (!BracketOperatorExpressionParser(szExpression, pos, value2, x, y, z, bracketIndex, true))
    {
        return false;
    }

    float value3;
    if (!BracketOperatorExpressionParser(szExpression, pos, value3, x, y, z, bracketIndex, true))
    {
        return false;
    }

    float value4;
    if (!BracketOperatorExpressionParser(szExpression, pos, value4, x, y, z, bracketIndex, false))
    {
        return false;
    }

    rst = operatorPtr->four_func(value1, value2, value3, value4);

    return true;
}

// --------------------------------------------------------------------------------------

bool MathExpressionParser(const char* szExpression, float& rst, float x, float y, float z)
{
    rst = 0.0f;
    if (!szExpression || !szExpression[0])
    {
        return false;
    }

    unsigned int lenExp = strlen(szExpression);
    unsigned int pos = 0;

    // ֵ
    float listValues[3];
    memset(listValues, 0, sizeof(listValues));
    unsigned int valuesCount = 0;

    // ׼Ŀ
    const Operator* standardSingleOperator = NULL;

    // ׼Ŀ
    const Operator* standardTwinOperators[2];
    memset(standardTwinOperators, 0, sizeof(standardTwinOperators));
    unsigned int operatorsCount = 0;

    standardSingleOperator = GetStandardSingleOperator(szExpression, pos);

    while (pos < lenExp)
    {
        if (valuesCount > operatorsCount)
        {
            const Operator* operatorPtr = GetStandardTwinOperator(szExpression, pos);
            if (!operatorPtr)
            {
                return false;
            }

            if (operatorsCount == 1)
            {
                if (operatorPtr->szName[0] == '+' || operatorPtr->szName[0] == '-')
                {
                    listValues[0] = standardTwinOperators[0]->twin_func(listValues[0], listValues[1]);
                    standardTwinOperators[0] = operatorPtr;
                    valuesCount--;
                }
                else
                {
                    standardTwinOperators[operatorsCount] = operatorPtr;
                    operatorsCount++;
                }
            }
            else
            {
                standardTwinOperators[operatorsCount] = operatorPtr;
                operatorsCount++;
            }
        }
        else
        {
            bool ableValue = false;
            float _value = 0.0f;

            // 
            ableValue = GetConstValue(_value, szExpression, pos);

            // ֵ
            if (!ableValue)
            {
                ableValue = GetIntegerValue(_value, szExpression, pos);
            }

            // ֵ
            if (!ableValue)
            {
                ableValue = GetFloatValue(_value, szExpression, pos);
            }

            // ֵ
            if (!ableValue)
            {
                char _param = GetParameter(szExpression, pos);
                if (_param == 'x')
                {
                    _value = x;
                    ableValue = true;
                }
                else if (_param == 'y')
                {
                    _value = y;
                    ableValue = true;
                }
                else if (_param == 'z')
                {
                    _value = z;
                    ableValue = true;
                }
            }

            // 
            if (!ableValue)
            {
                unsigned int bracketIndex;
                const Operator* operatorPtr = GetLeftBracketOperator(szExpression, pos, bracketIndex);
                if (operatorPtr)
                {
                    if (!BracketOperatorExpressionParser(szExpression, pos, _value, x, y, z, bracketIndex, false))
                    {
                        return false;
                    }
                    ableValue = true;
                }
            }

            // Ŀ
            if (!ableValue)
            {
                const Operator* operatorPtr = GetSingleOperator(szExpression, pos);
                if (operatorPtr)
                {
                    if (!SingleOperatorExpressionParser(szExpression, pos, operatorPtr, _value, x, y, z))
                    {
                        return false;
                    }
                    ableValue = true;
                }
            }

            // ˫Ŀ
            if (!ableValue)
            {
                const Operator* operatorPtr = GetTwinOperator(szExpression, pos);
                if (operatorPtr)
                {
                    if (!TwinOperatorExpressionParser(szExpression, pos, operatorPtr, _value, x, y, z))
                    {
                        return false;
                    }
                    ableValue = true;
                }
            }

            // Ŀ
            if (!ableValue)
            {
                const Operator* operatorPtr = GetThreeOperator(szExpression, pos);
                if (operatorPtr)
                {
                    if (!ThreeOperatorExpressionParser(szExpression, pos, operatorPtr, _value, x, y, z))
                    {
                        return false;
                    }
                    ableValue = true;
                }
            }

            // Ŀ
            if (!ableValue)
            {
                const Operator* operatorPtr = GetFourOperator(szExpression, pos);
                if (operatorPtr)
                {
                    if (!FourOperatorExpressionParser(szExpression, pos, operatorPtr, _value, x, y, z))
                    {
                        return false;
                    }
                    ableValue = true;
                }
            }

            if (!ableValue)
            {
                return false;
            }

            if (operatorsCount == 2)
            {
                assert(valuesCount == 2);
                listValues[valuesCount] = _value;
                const Operator* operatorPtr = standardTwinOperators[1];
                listValues[1] = operatorPtr->twin_func(listValues[1], listValues[2]);
                operatorsCount--;
            }
            else if (operatorsCount == 1)
            {
                assert(valuesCount == 1);
                listValues[valuesCount] = _value;
                const Operator* operatorPtr = standardTwinOperators[0];
                if (operatorPtr->szName[0] != '+' && operatorPtr->szName[0] != '-')
                {
                    listValues[0] = operatorPtr->twin_func(listValues[0], listValues[1]);
                    operatorsCount--;
                }
                else
                {
                    valuesCount++;
                }
            }
            else if (operatorsCount == 0)
            {
                assert(valuesCount == 0);
                listValues[valuesCount] = _value;

                if (standardSingleOperator)
                {
                    listValues[0] = standardSingleOperator->single_func(listValues[0]);
                    standardSingleOperator = NULL;
                }

                valuesCount++;
            }
            else
            {
                assert(0);
            }
        }
    }

    if (valuesCount != operatorsCount + 1)
    {
        return false;
    }

    if (operatorsCount == 1)
    {
        const Operator* operatorPtr = standardTwinOperators[0];
        listValues[0] = operatorPtr->twin_func(listValues[0], listValues[1]);
        operatorsCount--;
        valuesCount--;
    }

    rst = listValues[0];

    return true;
}

// --------------------------------------------------------------------------------------
