
/****************************************************************

  File name   :  MathExpressionParserBatch.cpp
  Author      :  Ҷ
  Version     :  1.0a
  Create Date :  2014/01/08 
  Description :  ѧʽ 
  Others      :  

*****************************************************************/

// --------------------------------------------------------------------------------------

#include <cstdlib>
#include <cstring>
#include <cmath>
#include <cstdio>
#include <assert.h>

#include "MathExpressionCommon.h"
#include "MathExpressionParserBatch.h"

// --------------------------------------------------------------------------------------

#pragma   warning(disable: 4996)

// --------------------------------------------------------------------------------------

struct ValueNode
{
    float value;
    float* valueList;
    unsigned int valueCount;

    ValueNode()
    {
        value = 0.0f;
        valueList = NULL;
        valueCount = 0;
    }

    ~ValueNode()
    {
        if (valueList)
        {
            free(valueList);
        }
    }

    void Clear()
    {
        value = 0.0f;
        if (valueList)
        {
            free(valueList);
            valueList = NULL;
        }
        valueCount = 0;
    }

    void Resize(unsigned int count)
    {
        if (valueCount == count)
        {
            return;
        }
        if (valueList)
        {
            free(valueList);
            valueList = 0;
        }
        if (count > 0)
        {
            valueList = (float*)malloc(sizeof(float)*count);
        }
        valueCount = count;
    }

    void SetValue(float v)
    {
        Clear();
        value = v;
    }

    void SetValue(const float* ptr, unsigned int count)
    {
        Resize(count);
        if (count)
        {
            memcpy(valueList, ptr, sizeof(float)*count);
        }
    }
};

bool OperatorCalculate(const Operator* operatorPtr, ValueNode* leftValuePtr, ValueNode* rightValuePtr, unsigned int count)
{
    if (!operatorPtr || !rightValuePtr)
    {
        return false;
    }

    if (operatorPtr->single_func)
    {
        if (rightValuePtr->valueList)
        {
            for (unsigned int i = 0; i < count; i++)
            {
                rightValuePtr->valueList[i] = operatorPtr->single_func(rightValuePtr->valueList[i]);
            }
        }
        else
        {
            rightValuePtr->value = operatorPtr->single_func(rightValuePtr->value);
        }

        return true;
    }
    else if (operatorPtr->twin_func && leftValuePtr)
    {
        if (leftValuePtr->valueList)
        {
            if (rightValuePtr->valueList)
            {
                for (unsigned int i = 0; i < count; i++)
                {
                    leftValuePtr->valueList[i] = operatorPtr->twin_func(leftValuePtr->valueList[i], rightValuePtr->valueList[i]);
                }
            }
            else
            {
                for (unsigned int i = 0; i < count; i++)
                {
                    leftValuePtr->valueList[i] = operatorPtr->twin_func(leftValuePtr->valueList[i], rightValuePtr->value);
                }
            }
        }
        else
        {
            if (rightValuePtr->valueList)
            {
                for (unsigned int i = 0; i < count; i++)
                {
                    rightValuePtr->valueList[i] = operatorPtr->twin_func(leftValuePtr->value, rightValuePtr->valueList[i]);
                    leftValuePtr->valueList = rightValuePtr->valueList;
                    rightValuePtr->valueList = NULL;
                }
            }
            else
            {
                leftValuePtr->value = operatorPtr->twin_func(leftValuePtr->value, rightValuePtr->value);
            }
        }

        return true;
    }
    else
    {
        return false;
    }
}

// --------------------------------------------------------------------------------------

bool MathExpressionParserBatch(const char* szExpression, unsigned int count, ValueNode& rst,
                               const float* xBufPtr, const float* yBufPtr, const float* zBufPtr);

// ڵӱʽ
bool BracketOperatorExpressionParser(const char* szExpression, unsigned int& pos, unsigned int count, ValueNode& rst,
                                     const float* xBufPtr, const float* yBufPtr, const float* zBufPtr, 
                                     unsigned int bracketIndex, bool endComma)
{
    rst.Clear();

    int nLeftBracket = 1;
    unsigned int bracketIndices[MAX_BRACKET_SIZE];
    bracketIndices[0] = bracketIndex;

    char szSubExp[SUB_EXPRESSION_MAX_LENGTH];
    memset(szSubExp, 0, SUB_EXPRESSION_MAX_LENGTH);

    unsigned int leftBI;
    unsigned int rightBI;

    unsigned int subExpLen = 0;
    while (szExpression[pos + subExpLen])
    {
        leftBI = IsLeftBracketLetter(szExpression[pos + subExpLen]);
        rightBI = IsLeftBracketLetter(szExpression[pos + subExpLen]);

        if (leftBI != -1)
        {
            if (nLeftBracket > MAX_BRACKET_SIZE - 1)
            {
                return false;
            }
            bracketIndices[nLeftBracket] = leftBI;
            nLeftBracket++;
        }
        else if (rightBI != -1)
        {
            nLeftBracket--;
            if (bracketIndices[nLeftBracket] != rightBI)
            {
                return false;
            }
        }
        else if (endComma && szExpression[pos + subExpLen] == ',')
        {
            // Զβ
            if (nLeftBracket == 1)
            {
                nLeftBracket--;
            }
            else
            {
                return false;
            }
        }

        if (nLeftBracket == 0)
        {
            break;
        }
        szSubExp[subExpLen] = szExpression[pos + subExpLen];
        subExpLen++;
    }

    if (nLeftBracket != 0)
    {
        return false;
    }

    pos += subExpLen + 1;

    return MathExpressionParserBatch(szSubExp, count, rst, xBufPtr, yBufPtr, zBufPtr);
}

// Ŀӱʽ
bool SingleOperatorExpressionParser(const char* szExpression, unsigned int& pos, const Operator* operatorPtr, 
                                    unsigned int count, ValueNode& rst,
                                    const float* xBufPtr, const float* yBufPtr, const float* zBufPtr)
{
    rst.Clear();

    if (!operatorPtr || !operatorPtr->single_func)
    {
        return false;
    }

    unsigned int bracketIndex;
    if (!GetLeftBracketOperator(szExpression, pos, bracketIndex))
    {
        return false;
    }

    if (!BracketOperatorExpressionParser(szExpression, pos, count, rst, xBufPtr, yBufPtr, zBufPtr, bracketIndex, false))
    {
        return false;
    }

    OperatorCalculate(operatorPtr, NULL, &rst, count);

    return true;
}

// ˫Ŀӱʽ
bool TwinOperatorExpressionParser(const char* szExpression, unsigned int& pos, const Operator* operatorPtr, 
                                  unsigned int count, ValueNode& rst,
                                  const float* xBufPtr, const float* yBufPtr, const float* zBufPtr)
{
    rst.Clear();

    if (!operatorPtr || !operatorPtr->twin_func)
    {
        return false;
    }

    unsigned int bracketIndex;
    if (!GetLeftBracketOperator(szExpression, pos, bracketIndex))
    {
        return false;
    }

    ValueNode& valueLeft = rst;
    if (!BracketOperatorExpressionParser(szExpression, pos, count, valueLeft, xBufPtr, yBufPtr, zBufPtr, bracketIndex, true))
    {
        return false;
    }

    ValueNode valueRight;
    if (!BracketOperatorExpressionParser(szExpression, pos, count, valueRight, xBufPtr, yBufPtr, zBufPtr, bracketIndex, false))
    {
        return false;
    }

    OperatorCalculate(operatorPtr, &valueLeft, &valueRight, count);

    return true;
}

bool MathExpressionParserBatch(const char* szExpression, unsigned int count, ValueNode& rst,
                               const float* xBufPtr, const float* yBufPtr, const float* zBufPtr)
{
    rst.Clear();

    if (!szExpression || !szExpression[0])
    {
        return false;
    }

    unsigned int lenExp = strlen(szExpression);
    unsigned int pos = 0;

    // ֵ
    ValueNode listValues[3];
    unsigned int valuesCount = 0;

    // ׼Ŀ
    const Operator* standardSingleOperator = NULL;

    // ׼Ŀ
    const Operator* standardTwinOperators[2];
    memset(standardTwinOperators, 0, sizeof(standardTwinOperators));
    unsigned int operatorsCount = 0;

    standardSingleOperator = GetStandardSingleOperator(szExpression, pos);

    while (pos < lenExp)
    {
        if (valuesCount > operatorsCount)
        {
            const Operator* operatorPtr = GetStandardTwinOperator(szExpression, pos);
            if (!operatorPtr)
            {
                return false;
            }

            if (operatorsCount == 1)
            {
                if (operatorPtr->szName[0] == '+' || operatorPtr->szName[0] == '-')
                {
                    OperatorCalculate(standardTwinOperators[0], &listValues[0], &listValues[1], count);
                    standardTwinOperators[0] = operatorPtr;
                    valuesCount--;
                }
                else
                {
                    standardTwinOperators[operatorsCount] = operatorPtr;
                    operatorsCount++;
                }
            }
            else
            {
                standardTwinOperators[operatorsCount] = operatorPtr;
                operatorsCount++;
            }
        }
        else
        {
            bool ableValue = false;
            ValueNode valueNode;

            // 
            ableValue = GetConstValue(valueNode.value, szExpression, pos);

            // ֵ
            if (!ableValue)
            {
                ableValue = GetIntegerValue(valueNode.value, szExpression, pos);
            }

            // ֵ
            if (!ableValue)
            {
                ableValue = GetFloatValue(valueNode.value, szExpression, pos);
            }

            // ֵ
            if (!ableValue)
            {
                char _param = GetParameter(szExpression, pos);
                if (_param == 'x')
                {
                    valueNode.SetValue(xBufPtr, count);
                    ableValue = true;
                }
                else if (_param == 'y')
                {
                    valueNode.SetValue(yBufPtr, count);
                    ableValue = true;
                }
                else if (_param == 'z')
                {
                    valueNode.SetValue(zBufPtr, count);
                    ableValue = true;
                }
            }

            // 
            if (!ableValue)
            {
                unsigned int bracketIndex;
                const Operator* operatorPtr = GetLeftBracketOperator(szExpression, pos, bracketIndex);
                if (operatorPtr)
                {
                    if (!BracketOperatorExpressionParser(szExpression, pos, count, valueNode, xBufPtr, yBufPtr, zBufPtr, bracketIndex, false))
                    {
                        return false;
                    }
                    ableValue = true;
                }
            }

            // Ŀ
            if (!ableValue)
            {
                const Operator* operatorPtr = GetSingleOperator(szExpression, pos);
                if (operatorPtr)
                {
                    if (!SingleOperatorExpressionParser(szExpression, pos, operatorPtr, count, valueNode, xBufPtr, yBufPtr, zBufPtr))
                    {
                        return false;
                    }
                    ableValue = true;
                }
            }

            // ˫Ŀ
            if (!ableValue)
            {
                const Operator* operatorPtr = GetTwinOperator(szExpression, pos);
                if (operatorPtr)
                {
                    if (!TwinOperatorExpressionParser(szExpression, pos, operatorPtr, count, valueNode, xBufPtr, yBufPtr, zBufPtr))
                    {
                        return false;
                    }
                    ableValue = true;
                }
            }

            if (!ableValue)
            {
                return false;
            }

            if (operatorsCount == 2)
            {
                assert(valuesCount == 2);

                // ͵ת
                listValues[valuesCount] = valueNode;
                valueNode.valueList = NULL;
                valueNode.valueCount = 0;

                const Operator* operatorPtr = standardTwinOperators[1];
                OperatorCalculate(operatorPtr, &listValues[1], &listValues[2], count);
                operatorsCount--;
            }
            else if (operatorsCount == 1)
            {
                assert(valuesCount == 1);

                // ͵ת
                listValues[valuesCount] = valueNode;
                valueNode.valueList = NULL;
                valueNode.valueCount = 0;

                const Operator* operatorPtr = standardTwinOperators[0];
                if (operatorPtr->szName[0] != '+' && operatorPtr->szName[0] != '-')
                {
                    OperatorCalculate(operatorPtr, &listValues[0], &listValues[1], count);
                    operatorsCount--;
                }
                else
                {
                    valuesCount++;
                }
            }
            else if (operatorsCount == 0)
            {
                assert(valuesCount == 0);

                // ͵ת
                listValues[valuesCount] = valueNode;
                valueNode.valueList = NULL;
                valueNode.valueCount = 0;

                if (standardSingleOperator)
                {
                    OperatorCalculate(standardSingleOperator, NULL, &listValues[0], count);
                    standardSingleOperator = NULL;
                }

                valuesCount++;
            }
            else
            {
                assert(0);
            }
        }
    }

    if (valuesCount != operatorsCount + 1)
    {
        return false;
    }

    if (operatorsCount == 1)
    {
        const Operator* operatorPtr = standardTwinOperators[0];
        OperatorCalculate(operatorPtr, &listValues[0], &listValues[1], count);
        operatorsCount--;
        valuesCount--;
    }

    // ͵ת
    rst = listValues[0];
    listValues[0].valueList = NULL;
    listValues[0].valueCount = 0;

    return true;
}

// --------------------------------------------------------------------------------------

bool MathExpressionParserBatch(const char* szExpression, unsigned int count, float* rstBufPtr,
                               const float* xBufPtr, const float* yBufPtr, const float* zBufPtr)
{
    ValueNode rst;
    if (!MathExpressionParserBatch(szExpression, count, rst, xBufPtr, yBufPtr, zBufPtr))
    {
        return false;
    }

    if (rst.valueList)
    {
        for (unsigned int i = 0; i < count; i++)
        {
            rstBufPtr[i] = rst.valueList[i];
        }
    }
    else
    {
        for (unsigned int i = 0; i < count; i++)
        {
            rstBufPtr[i] = rst.value;
        }
    }

    return false;
}